import logging
from typing import Tuple, List

import pandas as pd
import numpy as np

from carball.analysis.constants.field_constants import BALL_SIZE
from carball.generated.api import game_pb2
from carball.generated.api.player_pb2 import Player
from carball.generated.api.stats import events_pb2

logger = logging.getLogger(__name__)


# noinspection PyTypeChecker
class CarryData:
    def __init__(self, carry_frames: pd.DataFrame, start_frames: pd.Index, end_frames: pd.Index):
        self.carry_frames = carry_frames
        self.end_frames: List[int] = end_frames.values.tolist()
        self.start_frames: List[int] = start_frames.values.tolist()
        self.flicks = dict()

    def add_flick(self, carry_index: int, is_successful: bool):
        self.flicks[carry_index] = is_successful


class CarryDetection:

    def filter_frames(self, data_frame: pd.DataFrame) -> CarryData:
        valid_frames = data_frame[(data_frame.ball.pos_z > (BALL_SIZE + 5)) & (data_frame.ball.pos_z < 600)]

        return CarryData(valid_frames, *self.creat_start_end_frames(valid_frames))

    def create_carry_events(self, carry_data: CarryData, player: Player, proto_game: game_pb2.Game, data_frame: pd.DataFrame):
        """
        Gets all of the carry events and adds them to the protobuf for each individual player.
        """
        player_carry_data, xy_distance, player_frames = self.player_close_frames(carry_data.carry_frames,
                                                                                 carry_data.carry_frames[player.name])

        if len(player_carry_data.start_frames) == 0:
            # No dribbles no go
            return

        modified_carry_data = self.correct_carries(carry_data, player_carry_data, player, proto_game)

        self.add_carry_events(modified_carry_data, xy_distance, data_frame[player.name], player, proto_game)

    def player_close_frames(self, valid_frames: pd.DataFrame,
                            player_frames: pd.DataFrame) -> Tuple[CarryData, pd.DataFrame, pd.DataFrame]:
        """
        Filters out all frames that are not close enough to the player to be a valid carry.
        Filters out carries that are too short
        """

        null_values = player_frames.pos_x.isnull()
        true_valid_frames = valid_frames
        true_player_frames = player_frames
        if null_values.values.any():
            index = player_frames.index[null_values]
            true_valid_frames = valid_frames.drop(index=index)
            true_player_frames = player_frames.drop(index=index)

        xy_distance = ((true_player_frames.pos_x - true_valid_frames.ball.pos_x) ** 2 +
                       (true_player_frames.pos_y - true_valid_frames.ball.pos_y) ** 2) ** 0.5
        carry_frames = true_valid_frames[(xy_distance < (BALL_SIZE * 1.3)) &
                                         (true_player_frames.pos_z < true_valid_frames.ball.pos_z)]

        player_carry_data = CarryData(true_valid_frames, *self.creat_start_end_frames(carry_frames))
        return player_carry_data, xy_distance, true_player_frames
        # look up hits.
        # any hits by the same player within a continuous set of valid frames should count as carries

    def correct_carries(self, carry_data: CarryData, player_carry_data: CarryData,
                        player: Player, proto_game: game_pb2.Game) -> CarryData:
        """
        This modifies carries to correct for certain situations at a better granularity.
        """

        self.add_hits(carry_data, player_carry_data, player, proto_game.game_stats.hits)

        self.merge_carries(carry_data, player_carry_data)

        self.shorten_carries(player_carry_data, player, proto_game.game_stats.hits)

        # Look for the first carry frame and find all hits before it by the same player

        return player_carry_data

    def add_hits(self, carry_data: CarryData, player_carry_data: CarryData, player: Player, hit_list: List[events_pb2.Hit]):
        """
        Adds in frames to the carry in the beginning for the first hit of the carry.
        """
        hit_index = 0
        previous_end_frame = 0
        valid_frame_index = 0
        for frame_index in range(len(player_carry_data.start_frames)):

            starting_frame = player_carry_data.start_frames[frame_index]
            while (valid_frame_index < len(carry_data.end_frames) and
                   carry_data.end_frames[valid_frame_index] < starting_frame):
                valid_frame_index += 1
            valid_start_frame = carry_data.start_frames[valid_frame_index]

            # Get to the correct index before going backwards
            if hit_index >= len(hit_list):
                continue
            while hit_list[hit_index].frame_number < starting_frame:
                if hit_index == len(hit_list) - 1:
                    break
                hit_index += 1

            last_valid_hit = None
            while (hit_index >= 0 and hit_list[hit_index].player_id.id == player.id.id and
                   hit_list[hit_index].frame_number >= valid_start_frame and
                   hit_list[hit_index].frame_number > previous_end_frame):
                last_valid_hit = hit_list[hit_index]
                hit_index -= 1
            if last_valid_hit is not None:
                hit_index += 1

            if last_valid_hit is not None and last_valid_hit.frame_number < starting_frame:
                player_carry_data.start_frames[frame_index] = last_valid_hit.frame_number
            previous_end_frame = player_carry_data.end_frames[frame_index]

    def shorten_carries(self, player_carry_data: CarryData, player: Player, hit_list: List[events_pb2.Hit]):
        """
        Shortens carries in cases where the carry was ended by a player stealing the ball.
        """
        start_frames = player_carry_data.start_frames
        end_frames = player_carry_data.end_frames

        def valid_hit_number(hit_index, carry_index):
            return hit_list[hit_index].frame_number < end_frames[carry_index] and hit_index < len(hit_list) - 1

        hit_index = 0
        for carry_index in range(len(start_frames)):
            while hit_index < len(hit_list) - 1:
                if hit_list[hit_index].frame_number < start_frames[carry_index]:
                    hit_index += 1
                    continue
                if hit_list[hit_index].frame_number >= end_frames[carry_index]:
                    break

                invalid_hit = None
                last_player_hit = None
                while valid_hit_number(hit_index, carry_index):
                    if hit_list[hit_index].player_id.id != player.id.id:
                        # We need to potentially change how the carry occurred
                        if invalid_hit is None:
                            invalid_hit = hit_list[hit_index]
                        while hit_list[hit_index].player_id.id != player.id.id and valid_hit_number(hit_index,
                                                                                                    carry_index):
                            hit_index += 1
                        if hit_list[hit_index].frame_number >= end_frames[carry_index]:
                            # The player never hits it again, end the carry early.
                            player_carry_data.end_frames[carry_index] = invalid_hit.frame_number
                    else:
                        hit_index += 1
                    if (valid_hit_number(hit_index - 1, carry_index) and
                            hit_list[hit_index - 1].player_id.id == player.id.id):
                        last_player_hit = hit_list[hit_index - 1]
                if last_player_hit is None:
                    logger.error('The player never hit the ball during the "carry"')
                    end_frames[carry_index] = start_frames[carry_index]
                else:
                    most_recent_frame = max(last_player_hit.previous_hit_frame_number, start_frames[carry_index])
                    carry_frames = player_carry_data.carry_frames
                    dodge_data = carry_frames[player.name].dodge_active
                    dodge_data = dodge_data.loc[most_recent_frame:last_player_hit.frame_number]
                    has_flicked = dodge_data.where(dodge_data % 2 == 1).last_valid_index()
                    if has_flicked is not None:
                        ending = min(max(last_player_hit.frame_number + 10, end_frames[carry_index]), has_flicked + 20)
                        is_going_up = carry_frames.ball.pos_z[last_player_hit.frame_number:ending].is_monotonic
                        end_frames[carry_index] = has_flicked
                        if is_going_up:
                            player_carry_data.add_flick(carry_index, True)
                        else:
                            player_carry_data.add_flick(carry_index, False)
                    elif last_player_hit.frame_number > start_frames[carry_index]:
                        end_frames[carry_index] = last_player_hit.frame_number

            if hit_index >= len(hit_list) - 1:
                break

    def merge_carries(self, carry_data: CarryData, player_carry_data: CarryData):
        """
        Merges separated player carries that are still within the greater valid carry set.
        This modifies player_carry_data
        """
        start_frames = player_carry_data.start_frames
        end_frames = player_carry_data.end_frames

        merged_start = []
        merged_end = []

        if len(start_frames) == 0:
            return

        if len(start_frames) == 1:
            # Only one potential carry in the indexes, need to convert to mutable list
            player_carry_data.start_frames = [player_carry_data.start_frames[0]]
            player_carry_data.end_frames = [player_carry_data.end_frames[0]]
            return

        # This is all wrong need to redo merging.

        player_frame_index = 0
        for frame_index in range(len(carry_data.end_frames)):
            end_frame = carry_data.end_frames[frame_index]
            start_frame = carry_data.start_frames[frame_index]
            carry_end = None

            # if we have no more valid play indexes then we can exit
            if player_frame_index >= len(start_frames):
                break

            # if the first dribble is after a global carry already started then just continue
            if start_frames[player_frame_index] > end_frame:
                continue

            #  Increment till the player index is inside a valid start frame
            while player_frame_index < len(start_frames) and start_frames[player_frame_index] < start_frame:
                player_frame_index += 1

            # No more player dribbles
            if player_frame_index >= len(start_frames):
                break

            # if the first dribble is after a global carry already started then just continue
            if start_frames[player_frame_index] > end_frame:
                continue

            carry_start = start_frames[player_frame_index]
            # our start frame should now be inside a valid cary
            entered = False
            while player_frame_index < len(start_frames) and end_frames[player_frame_index] < end_frame:
                player_frame_index += 1
                entered = True

            # our end frame should now be outside a valid cary
            if entered:
                player_frame_index -= 1
            carry_end = end_frames[player_frame_index]

            if carry_end is not None:
                merged_start.append(carry_start)
                merged_end.append(carry_end)

        player_carry_data.start_frames = merged_start
        player_carry_data.end_frames = merged_end

    def creat_start_end_frames(self, carry_frames: pd.DataFrame) -> Tuple[List[int], List[int]]:
        """
        Returns the first and last frame of each potential carry
        """
        carry_frame_index = carry_frames.index
        shifted = carry_frame_index.to_series().shift(fill_value=0)
        neg_shifted = carry_frame_index.to_series().shift(periods=-1, fill_value=0)
        start_frames = carry_frame_index[abs(carry_frame_index - shifted) >= 2]
        end_frames = carry_frame_index[abs(carry_frame_index - neg_shifted) >= 2]
        return start_frames, end_frames

    def add_carry_events(self, carry_data: CarryData,
                         xy_distance: pd.DataFrame, player_frames: pd.DataFrame, player: Player, proto_game):

        start_frames = carry_data.start_frames
        end_frames = carry_data.end_frames
        for i in range(len(carry_data.start_frames)):
            total_time = carry_data.carry_frames.index.slice_locs(start_frames[i], end_frames[i])
            carry_frames = carry_data.carry_frames.iloc[total_time[0]:total_time[1]]
            total_time = carry_frames.game.delta.sum()
            if total_time < 1:
                continue

            ball_carry = proto_game.game_stats.ball_carries.add()
            ball_carry.start_frame_number = start_frames[i]
            ball_carry.end_frame_number = end_frames[i]
            ball_carry.player_id.id = player.id.id
            ball_carry.carry_time = total_time
            ball_carry.has_flick = i in carry_data.flicks and carry_data.flicks[i]
            ball_carry.straight_line_distance = self.get_straight_line_distance(player_frames,
                                                                                start_frames[i], end_frames[i])

            z_distance = carry_frames.ball.pos_z - player_frames.pos_z
            xy_carry = xy_distance[start_frames[i]:end_frames[i]]

            # adding stats for this carry
            carry_stats = ball_carry.carry_stats
            carry_stats.average_xy_distance = xy_carry.mean()
            carry_stats.average_z_distance = z_distance.mean()
            carry_stats.average_ball_z_velocity = carry_frames.ball.vel_z.mean()
            carry_stats.variance_xy_distance = xy_carry.var()
            carry_stats.variance_z_distance = z_distance.var()
            carry_stats.variance_ball_z_velocity = carry_frames.ball.vel_z.var()
            carry_stats.average_carry_speed = ((player_frames.vel_x ** 2 + player_frames.vel_y ** 2) ** 0.5).mean()
            carry_stats.distance_along_path = self.get_distance_along_path(player_frames,
                                                                           start_frames[i], end_frames[i])

    def get_distance_along_path(self, player_frames: pd.DataFrame, start_frame: int, end_frame: int):
        pos_x = player_frames.pos_x.loc[start_frame:end_frame].diff()
        pos_y = player_frames.pos_y.loc[start_frame:end_frame].diff()
        return ((pos_x ** 2 + pos_y ** 2) ** 0.5).sum()

    def get_straight_line_distance(self, player_frames: pd.DataFrame, start: int, end: int) -> pd.DataFrame:
        x1 = player_frames.pos_x[start]
        y1 = player_frames.pos_y[start]

        x2 = player_frames.pos_x[end]
        y2 = player_frames.pos_y[end]

        return ((x2 - x1) ** 2 + (y2 - y1) ** 2) ** 0.5



from .base import *
from carball.json_parser.dropshot import get_tile_mapping


class PlatformHandler(BaseActorHandler):

    def __init__(self, parser: object):
        super().__init__(parser)

    @classmethod
    def can_handle(cls, actor: dict) -> bool:
        return actor['TypeName'].startswith('ShatterShot_VFX.TheWorld:PersistentLevel.BreakOutActor_Platform_TA_')

    def update(self, actor: dict, frame_number: int, time: float, delta: float) -> None:
        tile_id = int(actor['TypeName'][actor['TypeName'].rfind('_') + 1:])

        tile_mapping = get_tile_mapping(self.parser.game.map)
        if tile_mapping is not None:
            tile_id = tile_mapping[tile_id]

        if tile_id not in self.parser.dropshot['tile_states']:
            self.parser.dropshot['tile_states'][tile_id] = 0

        damage_state = actor['TAGame.BreakOutActor_Platform_TA:DamageState']

        state = damage_state['tile_state']
        player_actor_id = damage_state['offender']
        tile_hit = damage_state['direct_hit']

        if state > self.parser.dropshot['tile_states'][tile_id]:

            if frame_number not in self.parser.dropshot['damage_events']:
                self.parser.dropshot['damage_events'][frame_number] = (player_actor_id, [])

            self.parser.dropshot['damage_events'][frame_number][1].append((tile_id, state, tile_hit))

        self.parser.dropshot['tile_states'][tile_id] = state

        if frame_number not in self.parser.dropshot['tile_frames']:
            self.parser.dropshot['tile_frames'][frame_number] = {}
        self.parser.dropshot['tile_frames'][frame_number][tile_id] = state
        self.parser.frames_data[frame_number][f'dropshot_tile_{tile_id}'] = state



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: api/stats/dropshot.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ...api import player_id_pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='api/stats/dropshot.proto',
  package='api.stats',
  serialized_pb=_b('\n\x18\x61pi/stats/dropshot.proto\x12\tapi.stats\x1a\x13\x61pi/player_id.proto\"?\n\x0c\x44ropshotTile\x12\n\n\x02id\x18\x01 \x01(\x05\x12#\n\x05state\x18\x02 \x01(\x0e\x32\x14.api.stats.TileState\":\n\x16\x44ropshotTileDamageStat\x12\n\n\x02id\x18\x01 \x01(\x05\x12\x14\n\x0ctotal_damage\x18\x02 \x01(\x05\"|\n\x11\x44ropshotTileStats\x12\x37\n\x0c\x64\x61mage_stats\x18\x01 \x03(\x0b\x32!.api.stats.DropshotTileDamageStat\x12\x15\n\rdamaged_tiles\x18\x02 \x01(\x05\x12\x17\n\x0f\x64\x65stroyed_tiles\x18\x03 \x01(\x05\"@\n\rDropshotStats\x12\x14\n\x0ctotal_damage\x18\x01 \x01(\x05\x12\x19\n\x11\x64\x61mage_efficiency\x18\x02 \x01(\x02\"T\n\x16\x44ropshotBallPhaseStats\x12\r\n\x05phase\x18\x01 \x01(\x05\x12\x0f\n\x07\x61verage\x18\x02 \x01(\x02\x12\x0b\n\x03max\x18\x03 \x01(\x02\x12\r\n\x05total\x18\x04 \x01(\x02\"\x87\x01\n\x13\x44ropshotDamageEvent\x12 \n\tplayer_id\x18\x01 \x01(\x0b\x32\r.api.PlayerId\x12\x14\n\x0c\x66rame_number\x18\x02 \x01(\x05\x12\x10\n\x08tile_hit\x18\x03 \x01(\x05\x12&\n\x05tiles\x18\x04 \x03(\x0b\x32\x17.api.stats.DropshotTile\"j\n\x1c\x44ropshotBallPhaseChangeEvent\x12\x14\n\x0c\x66rame_number\x18\x01 \x01(\x05\x12\x12\n\nball_phase\x18\x02 \x01(\x05\x12 \n\tplayer_id\x18\x03 \x01(\x0b\x32\r.api.PlayerId\"\xc0\x01\n\x11\x44ropshotGameStats\x12\x35\n\rdamage_events\x18\x01 \x03(\x0b\x32\x1e.api.stats.DropshotDamageEvent\x12\x42\n\x11\x62\x61ll_phase_events\x18\x02 \x03(\x0b\x32\'.api.stats.DropshotBallPhaseChangeEvent\x12\x30\n\ntile_stats\x18\x03 \x01(\x0b\x32\x1c.api.stats.DropshotTileStats*6\n\tTileState\x12\r\n\tUNDAMAGED\x10\x00\x12\x0b\n\x07\x44\x41MAGED\x10\x01\x12\r\n\tDESTROYED\x10\x02')
  ,
  dependencies=[player_id_pb2.DESCRIPTOR,])
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

_TILESTATE = _descriptor.EnumDescriptor(
  name='TileState',
  full_name='api.stats.TileState',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNDAMAGED', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DAMAGED', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DESTROYED', index=2, number=2,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=904,
  serialized_end=958,
)
_sym_db.RegisterEnumDescriptor(_TILESTATE)

TileState = enum_type_wrapper.EnumTypeWrapper(_TILESTATE)
UNDAMAGED = 0
DAMAGED = 1
DESTROYED = 2



_DROPSHOTTILE = _descriptor.Descriptor(
  name='DropshotTile',
  full_name='api.stats.DropshotTile',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='api.stats.DropshotTile.id', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='state', full_name='api.stats.DropshotTile.state', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=60,
  serialized_end=123,
)


_DROPSHOTTILEDAMAGESTAT = _descriptor.Descriptor(
  name='DropshotTileDamageStat',
  full_name='api.stats.DropshotTileDamageStat',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='api.stats.DropshotTileDamageStat.id', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='total_damage', full_name='api.stats.DropshotTileDamageStat.total_damage', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=125,
  serialized_end=183,
)


_DROPSHOTTILESTATS = _descriptor.Descriptor(
  name='DropshotTileStats',
  full_name='api.stats.DropshotTileStats',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='damage_stats', full_name='api.stats.DropshotTileStats.damage_stats', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='damaged_tiles', full_name='api.stats.DropshotTileStats.damaged_tiles', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='destroyed_tiles', full_name='api.stats.DropshotTileStats.destroyed_tiles', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=185,
  serialized_end=309,
)


_DROPSHOTSTATS = _descriptor.Descriptor(
  name='DropshotStats',
  full_name='api.stats.DropshotStats',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='total_damage', full_name='api.stats.DropshotStats.total_damage', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='damage_efficiency', full_name='api.stats.DropshotStats.damage_efficiency', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=311,
  serialized_end=375,
)


_DROPSHOTBALLPHASESTATS = _descriptor.Descriptor(
  name='DropshotBallPhaseStats',
  full_name='api.stats.DropshotBallPhaseStats',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='phase', full_name='api.stats.DropshotBallPhaseStats.phase', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='average', full_name='api.stats.DropshotBallPhaseStats.average', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='max', full_name='api.stats.DropshotBallPhaseStats.max', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='total', full_name='api.stats.DropshotBallPhaseStats.total', index=3,
      number=4, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=377,
  serialized_end=461,
)


_DROPSHOTDAMAGEEVENT = _descriptor.Descriptor(
  name='DropshotDamageEvent',
  full_name='api.stats.DropshotDamageEvent',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='player_id', full_name='api.stats.DropshotDamageEvent.player_id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='frame_number', full_name='api.stats.DropshotDamageEvent.frame_number', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='tile_hit', full_name='api.stats.DropshotDamageEvent.tile_hit', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='tiles', full_name='api.stats.DropshotDamageEvent.tiles', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=464,
  serialized_end=599,
)


_DROPSHOTBALLPHASECHANGEEVENT = _descriptor.Descriptor(
  name='DropshotBallPhaseChangeEvent',
  full_name='api.stats.DropshotBallPhaseChangeEvent',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='frame_number', full_name='api.stats.DropshotBallPhaseChangeEvent.frame_number', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='ball_phase', full_name='api.stats.DropshotBallPhaseChangeEvent.ball_phase', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='player_id', full_name='api.stats.DropshotBallPhaseChangeEvent.player_id', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=601,
  serialized_end=707,
)


_DROPSHOTGAMESTATS = _descriptor.Descriptor(
  name='DropshotGameStats',
  full_name='api.stats.DropshotGameStats',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='damage_events', full_name='api.stats.DropshotGameStats.damage_events', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='ball_phase_events', full_name='api.stats.DropshotGameStats.ball_phase_events', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='tile_stats', full_name='api.stats.DropshotGameStats.tile_stats', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=710,
  serialized_end=902,
)

_DROPSHOTTILE.fields_by_name['state'].enum_type = _TILESTATE
_DROPSHOTTILESTATS.fields_by_name['damage_stats'].message_type = _DROPSHOTTILEDAMAGESTAT
_DROPSHOTDAMAGEEVENT.fields_by_name['player_id'].message_type = player_id_pb2._PLAYERID
_DROPSHOTDAMAGEEVENT.fields_by_name['tiles'].message_type = _DROPSHOTTILE
_DROPSHOTBALLPHASECHANGEEVENT.fields_by_name['player_id'].message_type = player_id_pb2._PLAYERID
_DROPSHOTGAMESTATS.fields_by_name['damage_events'].message_type = _DROPSHOTDAMAGEEVENT
_DROPSHOTGAMESTATS.fields_by_name['ball_phase_events'].message_type = _DROPSHOTBALLPHASECHANGEEVENT
_DROPSHOTGAMESTATS.fields_by_name['tile_stats'].message_type = _DROPSHOTTILESTATS
DESCRIPTOR.message_types_by_name['DropshotTile'] = _DROPSHOTTILE
DESCRIPTOR.message_types_by_name['DropshotTileDamageStat'] = _DROPSHOTTILEDAMAGESTAT
DESCRIPTOR.message_types_by_name['DropshotTileStats'] = _DROPSHOTTILESTATS
DESCRIPTOR.message_types_by_name['DropshotStats'] = _DROPSHOTSTATS
DESCRIPTOR.message_types_by_name['DropshotBallPhaseStats'] = _DROPSHOTBALLPHASESTATS
DESCRIPTOR.message_types_by_name['DropshotDamageEvent'] = _DROPSHOTDAMAGEEVENT
DESCRIPTOR.message_types_by_name['DropshotBallPhaseChangeEvent'] = _DROPSHOTBALLPHASECHANGEEVENT
DESCRIPTOR.message_types_by_name['DropshotGameStats'] = _DROPSHOTGAMESTATS
DESCRIPTOR.enum_types_by_name['TileState'] = _TILESTATE

DropshotTile = _reflection.GeneratedProtocolMessageType('DropshotTile', (_message.Message,), dict(
  DESCRIPTOR = _DROPSHOTTILE,
  __module__ = 'api.stats.dropshot_pb2'
  # @@protoc_insertion_point(class_scope:api.stats.DropshotTile)
  ))
_sym_db.RegisterMessage(DropshotTile)

DropshotTileDamageStat = _reflection.GeneratedProtocolMessageType('DropshotTileDamageStat', (_message.Message,), dict(
  DESCRIPTOR = _DROPSHOTTILEDAMAGESTAT,
  __module__ = 'api.stats.dropshot_pb2'
  # @@protoc_insertion_point(class_scope:api.stats.DropshotTileDamageStat)
  ))
_sym_db.RegisterMessage(DropshotTileDamageStat)

DropshotTileStats = _reflection.GeneratedProtocolMessageType('DropshotTileStats', (_message.Message,), dict(
  DESCRIPTOR = _DROPSHOTTILESTATS,
  __module__ = 'api.stats.dropshot_pb2'
  # @@protoc_insertion_point(class_scope:api.stats.DropshotTileStats)
  ))
_sym_db.RegisterMessage(DropshotTileStats)

DropshotStats = _reflection.GeneratedProtocolMessageType('DropshotStats', (_message.Message,), dict(
  DESCRIPTOR = _DROPSHOTSTATS,
  __module__ = 'api.stats.dropshot_pb2'
  # @@protoc_insertion_point(class_scope:api.stats.DropshotStats)
  ))
_sym_db.RegisterMessage(DropshotStats)

DropshotBallPhaseStats = _reflection.GeneratedProtocolMessageType('DropshotBallPhaseStats', (_message.Message,), dict(
  DESCRIPTOR = _DROPSHOTBALLPHASESTATS,
  __module__ = 'api.stats.dropshot_pb2'
  # @@protoc_insertion_point(class_scope:api.stats.DropshotBallPhaseStats)
  ))
_sym_db.RegisterMessage(DropshotBallPhaseStats)

DropshotDamageEvent = _reflection.GeneratedProtocolMessageType('DropshotDamageEvent', (_message.Message,), dict(
  DESCRIPTOR = _DROPSHOTDAMAGEEVENT,
  __module__ = 'api.stats.dropshot_pb2'
  # @@protoc_insertion_point(class_scope:api.stats.DropshotDamageEvent)
  ))
_sym_db.RegisterMessage(DropshotDamageEvent)

DropshotBallPhaseChangeEvent = _reflection.GeneratedProtocolMessageType('DropshotBallPhaseChangeEvent', (_message.Message,), dict(
  DESCRIPTOR = _DROPSHOTBALLPHASECHANGEEVENT,
  __module__ = 'api.stats.dropshot_pb2'
  # @@protoc_insertion_point(class_scope:api.stats.DropshotBallPhaseChangeEvent)
  ))
_sym_db.RegisterMessage(DropshotBallPhaseChangeEvent)

DropshotGameStats = _reflection.GeneratedProtocolMessageType('DropshotGameStats', (_message.Message,), dict(
  DESCRIPTOR = _DROPSHOTGAMESTATS,
  __module__ = 'api.stats.dropshot_pb2'
  # @@protoc_insertion_point(class_scope:api.stats.DropshotGameStats)
  ))
_sym_db.RegisterMessage(DropshotGameStats)


# @@protoc_insertion_point(module_scope)



from typing import Dict

import pandas as pd

from ....analysis.stats.stats import BaseStat, HitStat
from ....generated.api import game_pb2
from ....generated.api.player_pb2 import Player
from ....generated.api.stats.events_pb2 import Hit
from ....generated.api.stats.team_stats_pb2 import TeamStats
from ....json_parser.game import Game


class PossessionStat(BaseStat, HitStat):

    def __init__(self):
        self.frame_possession_time_deltas = None

    def calculate_team_stat(self, team_stat_list: Dict[int, TeamStats], game: Game,
                            proto_game: game_pb2.Game, player_map: Dict[str, Player],
                            data_frame: pd.DataFrame):
        try:
            frame_possession_time_deltas = pd.concat(
                [
                    data_frame['ball', 'hit_team_no'],
                    data_frame['game', 'delta']
                ],
                axis=1
            )
        except KeyError:
            return
        frame_possession_time_deltas.columns = ['hit_team_no', 'delta']

        last_hit_possession = frame_possession_time_deltas.groupby('hit_team_no').sum()

        for index, stat in team_stat_list.items():
            try:
                stat.possession.possession_time = float(last_hit_possession.delta.loc[index])
            except KeyError:
                pass

    def initialize_hit_stat(self, game: Game, player_map: Dict[str, Player], data_frame: pd.DataFrame):
        try:
            self.frame_possession_time_deltas = pd.concat(
                [
                    data_frame['ball', 'hit_team_no'],
                    data_frame['game', 'delta']
                ],
                axis=1
            )
            self.frame_possession_time_deltas.columns = ['hit_team_no', 'delta']
        except KeyError:
            return

    def calculate_next_hit_stat(self, game: Game, proto_game: game_pb2.Game, saltie_hit: Hit, next_saltie_hit: Hit,
                                player_map: Dict[str, Player], hit_index: int):
        if self.frame_possession_time_deltas is None:
            return
        player = player_map[saltie_hit.player_id.id]
        next_player = player_map[next_saltie_hit.player_id.id]
        if player.is_orange == next_player.is_orange:
            hit_possession_time = self.frame_possession_time_deltas.delta.loc[
                                  saltie_hit.frame_number:next_saltie_hit.frame_number].sum()
            player.stats.possession.possession_time = player.stats.possession.possession_time + hit_possession_time
        else:
            hit_possession_time = self.frame_possession_time_deltas.delta.loc[
                                  saltie_hit.frame_number:next_saltie_hit.frame_number].sum()
            player.stats.possession.possession_time = player.stats.possession.possession_time + hit_possession_time
            proto_game.game_stats.neutral_possession_time = proto_game.game_stats.neutral_possession_time + hit_possession_time



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: api/stats/ball_stats.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ...api.stats import stats_pb2
from ...api.stats import extra_mode_stats_pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='api/stats/ball_stats.proto',
  package='api.stats',
  serialized_pb=_b('\n\x1a\x61pi/stats/ball_stats.proto\x12\tapi.stats\x1a\x15\x61pi/stats/stats.proto\x1a api/stats/extra_mode_stats.proto\"\xc6\x01\n\tBallStats\x12\x1f\n\x05speed\x18\x01 \x01(\x0b\x32\x10.api.stats.Speed\x12>\n\x15positional_tendencies\x18\x02 \x01(\x0b\x32\x1f.api.stats.PositionalTendencies\x12%\n\x08\x61verages\x18\x03 \x01(\x0b\x32\x13.api.stats.Averages\x12\x31\n\nextra_mode\x18\x04 \x01(\x0b\x32\x1d.api.stats.ExtraModeBallStats')
  ,
  dependencies=[stats_pb2.DESCRIPTOR,extra_mode_stats_pb2.DESCRIPTOR,])
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_BALLSTATS = _descriptor.Descriptor(
  name='BallStats',
  full_name='api.stats.BallStats',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='speed', full_name='api.stats.BallStats.speed', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='positional_tendencies', full_name='api.stats.BallStats.positional_tendencies', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='averages', full_name='api.stats.BallStats.averages', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='extra_mode', full_name='api.stats.BallStats.extra_mode', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=99,
  serialized_end=297,
)

_BALLSTATS.fields_by_name['speed'].message_type = stats_pb2._SPEED
_BALLSTATS.fields_by_name['positional_tendencies'].message_type = stats_pb2._POSITIONALTENDENCIES
_BALLSTATS.fields_by_name['averages'].message_type = stats_pb2._AVERAGES
_BALLSTATS.fields_by_name['extra_mode'].message_type = extra_mode_stats_pb2._EXTRAMODEBALLSTATS
DESCRIPTOR.message_types_by_name['BallStats'] = _BALLSTATS

BallStats = _reflection.GeneratedProtocolMessageType('BallStats', (_message.Message,), dict(
  DESCRIPTOR = _BALLSTATS,
  __module__ = 'api.stats.ball_stats_pb2'
  # @@protoc_insertion_point(class_scope:api.stats.BallStats)
  ))
_sym_db.RegisterMessage(BallStats)


# @@protoc_insertion_point(module_scope)



import logging
from typing import Dict

import numpy as np
import pandas as pd
from carball.analysis.constants.field_constants import FieldConstants

from carball.analysis.stats.utils.pandas_utils import sum_deltas_by_truthy_data
from ....analysis.stats.stats import BaseStat
from ....generated.api import game_pb2
from ....generated.api.player_pb2 import Player
from ....generated.api.stats.player_stats_pb2 import PlayerStats
from ....json_parser.actor.boost import BOOST_PER_SECOND
from ....json_parser.game import Game

logger = logging.getLogger(__name__)


class BoostStat(BaseStat):
    field_constants = FieldConstants()

    def calculate_player_stat(self, player_stat_map: Dict[str, PlayerStats], game: Game, proto_game: game_pb2.Game,
                              player_map: Dict[str, Player], data_frame: pd.DataFrame):
        for player_key, stats in player_stat_map.items():

            proto_boost = stats.boost
            player_name = player_map[player_key].name
            player_data_frame = data_frame[player_name].copy()
            player_data_frame.loc[:, 'delta'] = data_frame['game'].delta
            proto_boost.boost_usage = self.get_player_boost_usage(player_data_frame)

            proto_boost.wasted_usage = self.get_player_boost_usage_max_speed(player_data_frame)

            proto_boost.time_full_boost = self.get_time_with_max_boost(data_frame, player_data_frame)
            proto_boost.time_low_boost = self.get_time_with_low_boost(data_frame, player_data_frame)
            proto_boost.time_no_boost = self.get_time_with_zero_boost(data_frame, player_data_frame)
            proto_boost.average_boost_level = self.get_average_boost_level(player_data_frame)

            if 'boost_collect' not in player_data_frame:
                logger.warning('%s did not collect any boost', player_key)
            else:
                gains_index = player_data_frame['boost'].diff().clip(0)
                gains_index = gains_index.loc[gains_index > 0].index.to_numpy()
                collect_frames = player_data_frame.loc[player_data_frame.index[player_data_frame['boost_collect'] > 34]]
                # Have to loop to fuzzy match
                wasted_big = 0
                for index in collect_frames.index.to_numpy():
                    idx = gains_index[(np.abs(gains_index - index).argmin())]
                    int_idx = player_data_frame.index.get_loc(idx)
                    wasted_big += player_data_frame['boost'].iloc[int_idx - 1] / 256 * 100

                collect_frames = player_data_frame.loc[player_data_frame.index[player_data_frame['boost_collect'] <= 34]]
                prior_vals = np.empty([0])
                for index in collect_frames.index.to_numpy():
                    idx = gains_index[(np.abs(gains_index - index).argmin())]
                    int_idx = player_data_frame.index.get_loc(idx)
                    val = player_data_frame['boost'].iloc[int_idx-1]
                    prior_vals = np.append(prior_vals, val)
                deltas = ((prior_vals + 30.6) - 255)
                wasted_small = deltas[deltas > 0].sum() / 256 * 100

                collection = self.get_player_boost_collection(player_data_frame)
                proto_boost.wasted_collection = wasted_big + wasted_small
                proto_boost.wasted_big = wasted_big
                proto_boost.wasted_small = wasted_small

                if 'small' in collection and collection['small'] is not None:
                    proto_boost.num_small_boosts = collection['small']
                if 'big' in collection and collection['big'] is not None:
                    proto_boost.num_large_boosts = collection['big']

                proto_boost.num_stolen_boosts = self.get_num_stolen_boosts(player_data_frame,
                                                                           player_map[player_key].is_orange)

    @staticmethod
    def get_player_boost_usage(player_dataframe: pd.DataFrame) -> np.float64:
        return (BOOST_PER_SECOND * (player_dataframe.delta * player_dataframe.boost_active)).sum() / 255 * 100
        # _diff = -player_dataframe.boost.diff()
        # boost_usage = _diff[_diff > 0].sum() / 255 * 100
        # return boost_usage

    @staticmethod
    def get_average_boost_level(player_dataframe: pd.DataFrame) -> np.float64:
        return player_dataframe.boost.mean(skipna=True) / 255 * 100

    @classmethod
    def get_num_stolen_boosts(cls, player_dataframe: pd.DataFrame, is_orange):
        big = cls.field_constants.get_big_pads()
        # Get big pads below or above 0 depending on team
        # The index of y position is 1. The index of the label is 2.
        if is_orange:
            opponent_pad_labels = big[big[:, 1] < 0][:, 2] #big[where[y] is < 0][labels]
        else:
            opponent_pad_labels = big[big[:, 1] > 0][:, 2] #big[where[y] is > 0][labels]
        # Count all the places where isin = True by summing
        stolen = player_dataframe.boost_collect.isin(opponent_pad_labels).sum()
        return stolen

    @staticmethod
    def get_player_boost_usage_max_speed(player_dataframe: pd.DataFrame) -> np.float64:
        _diff = -player_dataframe.boost.diff()

        speed: pd.Series = (player_dataframe.vel_x ** 2 +
                            player_dataframe.vel_y ** 2 +
                            player_dataframe.vel_z ** 2) ** 0.5

        _diff = _diff.rename('boost')
        speed = speed.rename('speed')

        combined = pd.concat([_diff, speed], axis=1)

        wasted_boost = combined[(combined.speed > 22000) & (combined.boost > 0) & (combined.boost < 10)]
        boost_usage = wasted_boost.boost.sum() / 255 * 100
        return boost_usage

    @staticmethod
    def get_time_with_max_boost(data_frame: pd.DataFrame, player_dataframe: pd.DataFrame) -> np.float64:
        return sum_deltas_by_truthy_data(data_frame, player_dataframe.boost > 252.45)  # 100% visible boost

    @staticmethod
    def get_time_with_low_boost(data_frame: pd.DataFrame, player_dataframe: pd.DataFrame) -> np.float64:  # less than 25
        return sum_deltas_by_truthy_data(data_frame, player_dataframe.boost < 63.75)  # 25 / 100 * 255

    @staticmethod
    def get_time_with_zero_boost(data_frame: pd.DataFrame, player_dataframe: pd.DataFrame) -> np.float64:  # at 0 boost
        return sum_deltas_by_truthy_data(data_frame, player_dataframe.boost == 0)

    @staticmethod
    def get_player_boost_collection(player_dataframe: pd.DataFrame) -> Dict[str, int]:
        try:
            big_counts = (player_dataframe['boost_collect'] > 34).sum()
            small_counts = (player_dataframe['boost_collect'] <= 34).sum()
            ret = {}
            if big_counts > 0:
                ret['big'] = big_counts
            if small_counts > 0:
                ret['small'] = small_counts
        except (AttributeError, KeyError):
            return {}
        return ret



import logging
from typing import Dict

import pandas as pd
from ....analysis.stats.stats import BaseStat
from ....generated.api import game_pb2
from ....generated.api.player_pb2 import Player
from ....generated.api.stats.player_stats_pb2 import PlayerStats
from ....json_parser.game import Game

logger = logging.getLogger(__name__)


class CarryStat(BaseStat):
    def calculate_player_stat(self, player_stat_map: Dict[str, PlayerStats], game: Game, proto_game: game_pb2.Game,
                              player_map: Dict[str, Player], data_frame: pd.DataFrame):
        for event_carry in proto_game.game_stats.ball_carries:
            player_stats = player_stat_map[event_carry.player_id.id]
            player_carry_stats = player_stats.ball_carries
            player_carry_stats.total_carries += 1
            if event_carry.has_flick:
                player_carry_stats.total_flicks += 1

            # time
            player_carry_stats.total_carry_time += event_carry.carry_time
            if player_carry_stats.longest_carry < event_carry.carry_time:
                player_carry_stats.longest_carry = event_carry.carry_time

            # distance
            player_carry_stats.total_carry_distance += event_carry.straight_line_distance
            if player_carry_stats.furthest_carry < event_carry.straight_line_distance:
                player_carry_stats.furthest_carry = event_carry.straight_line_distance

            if player_carry_stats.fastest_carry_speed < event_carry.carry_stats.average_carry_speed:
                player_carry_stats.fastest_carry_speed = event_carry.carry_stats.average_carry_speed
            self.add_ball_stats(player_carry_stats.carry_stats, event_carry.carry_stats)

        for player_key in player_stat_map:
            player_carry_stats = player_stat_map[player_key].ball_carries
            num_carries = player_carry_stats.total_carries
            if num_carries == 0:
                continue
            player_carry_stats.average_carry_time = player_carry_stats.total_carry_time / num_carries
            self.average_ball_stats(player_carry_stats.carry_stats, num_carries)

    def add_ball_stats(self, player_carry, event_carry):
        player_carry.average_z_distance += event_carry.average_z_distance
        player_carry.average_xy_distance += event_carry.average_xy_distance
        player_carry.average_ball_z_velocity += event_carry.average_ball_z_velocity
        player_carry.variance_xy_distance += event_carry.variance_xy_distance
        player_carry.variance_z_distance += event_carry.variance_z_distance
        player_carry.variance_ball_z_velocity += event_carry.variance_ball_z_velocity
        player_carry.average_carry_speed += event_carry.average_carry_speed
        player_carry.distance_along_path += event_carry.distance_along_path

    def average_ball_stats(self, player_carry, num_carries):
        player_carry.average_z_distance /= num_carries
        player_carry.average_xy_distance /= num_carries
        player_carry.average_ball_z_velocity /= num_carries
        player_carry.variance_xy_distance /= num_carries
        player_carry.variance_z_distance /= num_carries
        player_carry.variance_ball_z_velocity /= num_carries
        player_carry.average_carry_speed /= num_carries
        player_carry.distance_along_path /= num_carries



from typing import Dict

import pandas as pd

from ....analysis.stats.utils.pandas_utils import sum_deltas_by_truthy_data
from ....analysis.stats.stats import BaseStat
from ....generated.api import game_pb2
from ....generated.api.player_pb2 import Player
from ....generated.api.stats.player_stats_pb2 import PlayerStats
from ....json_parser.game import Game


class SpeedTendencies(BaseStat):

    def calculate_stat(self, proto_stat, game: Game, proto_game: game_pb2.Game, player_map: Dict[str, Player],
                       data_frame: pd.DataFrame):

        ball = data_frame['ball']
        if 'vel_x' not in ball:
            return
        speed: pd.Series = (ball.vel_x ** 2 +
                            ball.vel_y ** 2 +
                            ball.vel_z ** 2) ** 0.5

        average_speed = speed.mean()

        proto_stat.ball_stats.averages.average_speed = average_speed

    def calculate_player_stat(self, player_stat_map: Dict[str, PlayerStats], game: Game, proto_game: game_pb2.Game,
                              player_map: Dict[str, Player], data_frame: pd.DataFrame):

        for key, player in player_map.items():
            self.calculate_speed_for_data_frame(player, data_frame)

    @classmethod
    def calculate_speed_for_data_frame(cls, player: Player, data_frame: pd.DataFrame):
        player_data_frame = data_frame[player.name]

        speed: pd.Series = (player_data_frame.vel_x ** 2 +
                            player_data_frame.vel_y ** 2 +
                            player_data_frame.vel_z ** 2) ** 0.5

        average_speed = speed.mean()

        player.stats.averages.average_speed = average_speed

        player.stats.speed.time_at_super_sonic = sum_deltas_by_truthy_data(data_frame, speed >= 22000)
        player.stats.speed.time_at_slow_speed = sum_deltas_by_truthy_data(data_frame, speed <= 7000)
        player.stats.speed.time_at_boost_speed = sum_deltas_by_truthy_data(data_frame, speed > 14100)




import os

from ... import analyze_replay_file
from ..utils import get_replay_path


# All relevant files begin with df_ .

def create_df_docs():
    replay_path = get_replay_path("SHORT_SAMPLE.replay")

    working_dir = os.path.dirname(__file__)

    df_summary_path = os.path.join(working_dir, 'df_summary.md')
    df_methods_path = os.path.join(working_dir, 'df_methods.md')
    df_docs_path = os.path.join(working_dir, "data_frame.md")

    analysis = analyze_replay_file(replay_path, clean=False)

    data_frame = analysis.get_data_frame()

    df_docs = open(df_docs_path, "w")
    df_docs.write("This file is autogenerated do not edit\n")
    df_docs.write("# pandas.DataFrame\n")

    with open(df_summary_path) as summary:
        for line in summary:
            df_docs.write(line)

    df_docs.write("\n")

    with open(df_methods_path) as methods:
        for line in methods:
            df_docs.write(line)

    df_docs.write("\n")

    df_docs.write("## Example (list of columns)\n")
    df_docs.write("Each subheading is the primary column heading, and each row is the secondary column heading."
                  "The 'Player' subheading will be unique for each player in the replay (i.e. for a 3v3 game, there "
                  "will be 6 distinct players)\n")

    current_heading = ""
    for c in data_frame.columns:
        if current_heading != c[0]:
            current_heading = c[0]
            df_docs.write("\n#### " + c[0])

        df_docs.write("\n\t" + c[1])


def test_benchmark(benchmark):
    benchmark(create_df_docs)


if __name__ == '__main__':
    create_df_docs()



import gzip
import os

from ..analysis.utils.pandas_manager import PandasManager

OUTPUT_DIR = os.path.join('..', 'replays', 'pickled')
OUTPUT_FORMAT = '.gzip'
BASE_DIR = os.path.dirname(os.path.dirname(__file__))


class Test_Database():
    def setup_method(self):
        self.pandas = None
        if not os.path.isdir(OUTPUT_DIR):
            os.makedirs(OUTPUT_DIR)

        for filename in [f for f in os.listdir(OUTPUT_DIR) if os.path.isfile(os.path.join(OUTPUT_DIR, f))]:
            if OUTPUT_FORMAT in filename:
                filepath = os.path.join(OUTPUT_DIR, filename)
                with gzip.open(filepath, 'rb') as file:
                    self.pandas = PandasManager.read_numpy_from_memory(file)

    def test_replay_attrs(self):
        if self.pandas is not None:
            print(self.pandas)



from enum import Enum


class CheckErrorLevel(Enum):
    MINOR = 1
    MAJOR = 2
    CRITICAL = 3
    IGNORE_ERRORS = 99999


class CheckError(Exception):
    def __init__(self, level: CheckErrorLevel, message: str):
        self.level = level
        self.message = message



import logging

import pandas as pd

from ...generated.api import game_pb2
from ...json_parser.game import Game


logger = logging.getLogger(__name__)


def remove_invalid_players(game: Game, data_frame: pd.DataFrame, proto_game: game_pb2.Game, player_map):
    """
    Finds and removes invalid players from all relevant fields.

    :param game: The Game object.
    :param data_frame: The pandas.DataFrame object.
    :param proto_game: The game's protobuf object.
    :param player_map: The map of players' online IDs to the Player objects.
    """

    # Get invalid players; if none - return.
    invalid_players = get_invalid_players(proto_game, data_frame)
    if len(invalid_players) == 0:
        return

    # Remove player from Game and game_pb2.Game
    game.players = [player for player in game.players if player.name not in invalid_players]
    remove_player_from_protobuf(proto_game, player_map, invalid_players)

    logger.warning("Invalid player(s) removed: " + str(invalid_players))


def get_invalid_players(proto_game: game_pb2.Game, data_frame: pd.DataFrame):
    """
    :return: An array of invalid players' names.
    """

    invalid_players = []
    for player in proto_game.players:
        name = player.name
        if player.time_in_game < 5 or name not in data_frame or ('pos_x' not in data_frame[name]):
            invalid_players.append(player.name)

    return invalid_players


def remove_player_from_protobuf(proto_game: game_pb2.Game, player_map, invalid_players):
    """
    Removes invalid players from all relevant proto_game fields.
    """

    proto_players = proto_game.players
    i = 0
    total_length = len(proto_players)

    while i < total_length:
        player = proto_players[i]
        if player.name in invalid_players:
            del player_map[player.id.id]
            del proto_players[i]

            remove_player_from_team(proto_game, player)

            i -= 1
            total_length -= 1
        i += 1


def remove_player_from_team(proto_game: game_pb2.Game, player):
    """
    Removes given player from proto_game.teams
    """

    for team in proto_game.teams:
        for i in range(len(team.player_ids)):
            if team.player_ids[i] == player.id:
                del team.player_ids[i]
                break



from carball.json_parser.game import Game
from carball.generated.api import game_pb2
from carball.generated.api.metadata.game_metadata_pb2 import RANKED_DROPSHOT, UNRANKED_DROPSHOT


def create_dropshot_damage_events(game: Game, proto_game: game_pb2.Game):
    if game.game_info.playlist not in [RANKED_DROPSHOT, UNRANKED_DROPSHOT] and game.map != 'ShatterShot_P':
        return

    proto_damages = proto_game.game_stats.dropshot_stats.damage_events

    for event in game.dropshot['damage_events']:
        proto_damage_event = proto_damages.add()
        proto_damage_event.frame_number = event['frame_number']
        proto_damage_event.player_id.id = str(event['player'].online_id)

        for tile in event['tiles']:
            proto_tile = proto_damage_event.tiles.add()
            proto_tile.id = tile[0]
            proto_tile.state = tile[1]

            if tile[2]:
                proto_damage_event.tile_hit = tile[0]



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: api/player_id.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='api/player_id.proto',
  package='api',
  serialized_pb=_b('\n\x13\x61pi/player_id.proto\x12\x03\x61pi\"\x16\n\x08PlayerId\x12\n\n\x02id\x18\x01 \x01(\t')
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_PLAYERID = _descriptor.Descriptor(
  name='PlayerId',
  full_name='api.PlayerId',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='api.PlayerId.id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=28,
  serialized_end=50,
)

DESCRIPTOR.message_types_by_name['PlayerId'] = _PLAYERID

PlayerId = _reflection.GeneratedProtocolMessageType('PlayerId', (_message.Message,), dict(
  DESCRIPTOR = _PLAYERID,
  __module__ = 'api.player_id_pb2'
  # @@protoc_insertion_point(class_scope:api.PlayerId)
  ))
_sym_db.RegisterMessage(PlayerId)


# @@protoc_insertion_point(module_scope)



import re
from .base import *


class RumbleItemHandler(BaseActorHandler):

    @classmethod
    def can_handle(cls, actor: dict) -> bool:
        return actor['TypeName'].startswith('Archetypes.SpecialPickups.SpecialPickup_')

    def update(self, actor: dict, frame_number: int, time: float, delta: float) -> None:
        car_actor_id = actor.get('TAGame.CarComponent_TA:Vehicle', {}).get('actor', None)
        if car_actor_id is not None and car_actor_id in self.parser.current_car_ids_to_collect:
            player_actor_id = self.parser.car_player_ids[car_actor_id]
            item_name = actor['TypeName'] \
                .replace('Archetypes.SpecialPickups.SpecialPickup_', '')
            # CamelCase to snake_case
            item_name = re.sub('([A-Z]+)', r'_\1', item_name).lower()[1:]
            # item is active when this is odd
            item_active = actor.get(COMPONENT_REPLICATED_ACTIVE_KEY, 0) % 2 == 1

            self.parser.player_data[player_actor_id][frame_number]['power_up'] = item_name
            self.parser.player_data[player_actor_id][frame_number]['power_up_active'] = item_active



from .base import *

BOOST_PER_SECOND = 80 * 1 / .93  # boost used per second out of 255
REPLICATED_PICKUP_KEY = 'TAGame.VehiclePickup_TA:ReplicatedPickupData'
REPLICATED_PICKUP_KEY_168 = 'TAGame.VehiclePickup_TA:NewReplicatedPickupData'


def get_boost_actor_data(actor: dict):
    if REPLICATED_PICKUP_KEY in actor:
        actor = actor[REPLICATED_PICKUP_KEY]
        if actor is not None and actor != -1:
            actor = actor['pickup']
            if actor is not None and 'instigator_id' in actor and actor["instigator_id"] != -1:
                return actor
    elif REPLICATED_PICKUP_KEY_168 in actor:
        actor = actor[REPLICATED_PICKUP_KEY_168]
        if actor is not None and actor != -1:
            actor = actor['pickup_new']
            if actor is not None and 'instigator_id' in actor and actor["instigator_id"] != -1:
                return actor
    return None


class BoostHandler(BaseActorHandler):
    type_name = 'Archetypes.CarComponents.CarComponent_Boost'

    def update(self, actor: dict, frame_number: int, time: float, delta: float) -> None:
        car_actor_id = actor.get('TAGame.CarComponent_TA:Vehicle', {}).get('actor', None)

        if car_actor_id is None or car_actor_id not in self.parser.current_car_ids_to_collect:
            return

        player_actor_id = self.parser.car_player_ids[car_actor_id]
        boost_is_active_random_int = actor.get(
            COMPONENT_ACTIVE_KEY,
            actor.get(COMPONENT_REPLICATED_ACTIVE_KEY, False))
        # boost_is_active when random_int is odd?!
        boost_is_active = (boost_is_active_random_int % 2 == 1)
        if boost_is_active:
            # manually decrease car boost amount (not shown in replay)
            # i assume game calculates the decrease itself similarly
            boost_amount = max(0, actor.get('TAGame.CarComponent_Boost_TA:ReplicatedBoostAmount',
                                            0) - delta * BOOST_PER_SECOND)
            actor['TAGame.CarComponent_Boost_TA:ReplicatedBoostAmount'] = boost_amount
        else:
            boost_amount = actor.get('TAGame.CarComponent_Boost_TA:ReplicatedBoostAmount', None)

        self.parser.player_data[player_actor_id][frame_number]['boost'] = boost_amount
        self.parser.player_data[player_actor_id][frame_number]['boost_active'] = boost_is_active


class BoostPickupHandler(BaseActorHandler):

    @classmethod
    def can_handle(cls, actor: dict) -> bool:
        return actor['ClassName'] == 'TAGame.VehiclePickup_Boost_TA'

    def update(self, actor: dict, frame_number: int, time: float, delta: float) -> None:
        boost_actor = get_boost_actor_data(actor)
        if boost_actor is not None:
            car_actor_id = boost_actor['instigator_id']
            if car_actor_id in self.parser.car_player_ids:
                player_actor_id = self.parser.car_player_ids[car_actor_id]
                if frame_number in self.parser.player_data[player_actor_id]:
                    actor = self.parser.player_data[player_actor_id]
                    frame_number_look_back = frame_number - 1
                    previous_boost_data = None
                    while frame_number_look_back >= 0:
                        try:
                            previous_boost_data = actor[frame_number_look_back]['boost']
                        except KeyError:
                            previous_boost_data = None
                        if previous_boost_data is not None:
                            break
                        frame_number_look_back -= 1
                    try:
                        current_boost_data = actor[frame_number]['boost']
                    except KeyError:
                        current_boost_data = None

                    # Ignore any phantom boosts
                    if (previous_boost_data is not None and current_boost_data is not None and
                            (255 > previous_boost_data < current_boost_data)):
                        actor[frame_number]['boost_collect'] = True
                        # set to false after acknowledging it's turned True
                        # it does not turn back false immediately although boost is only collected once.
                        # using actor_id!=-1
                        boost_actor["instigator_id"] = -1



from .game_info import GameInfoHandler
from .game_event import GameEventHandler
from .ball import BallHandler
from .player import PlayerHandler
from .team import TeamHandler
from .car import CarHandler
from .jump import JumpHandler, DodgeHandler, DoubleJumpHandler
from .boost import BoostHandler, BoostPickupHandler
from .camera import CameraSettingsHandler
from .rumble import RumbleItemHandler
from .dropshot import PlatformHandler



from carball.generated.api import game_pb2


class ProtobufManager:

    @staticmethod
    def write_proto_out_to_file(file, proto_game: game_pb2.Game):
        file.write(proto_game.SerializeToString())

    @staticmethod
    def read_proto_out_from_file(file):
        proto_game = game_pb2.Game()
        proto_game.ParseFromString(file.read())
        return proto_game



import logging
from typing import Dict
from carball.json_parser.game import Game
from carball.generated.api import game_pb2
from carball.generated.api.game_pb2 import mutators_pb2 as mutators
from carball.generated.api.player_pb2 import Player

log = logging.getLogger(__name__)


def create_dropshot_ball_events(game: Game, proto_game: game_pb2.Game, player_map: Dict[str, Player]):
    if game.ball_type != mutators.BREAKOUT:
        return

    hits = list(proto_game.game_stats.hits)
    proto_events = proto_game.game_stats.dropshot_stats.ball_phase_events
    for event in game.dropshot['ball_events']:

        frame_number = event['frame_number']
        proto_event = proto_events.add()
        proto_event.frame_number = frame_number
        proto_event.ball_phase = event['state']

        while len(hits) > 1 and hits[1].frame_number <= frame_number:
            hits.pop(0)

        hit = hits.pop(0)

        if hit.frame_number != frame_number:
            log.warning(f'Did not find exact hit event for dropshot ball event at frame {frame_number}, hit frame {hit.frame_number}')

        player = player_map[hit.player_id.id]

        if player.is_orange != bool(event['team']):
            log.warning(f'Team does not match in dropshot ball event ({frame_number}) and hit event ({hit.frame_number})')
            return

        proto_event.player_id.id = player.id.id



import json
import logging
import re
from datetime import datetime
from typing import List

import pandas as pd

from .goal import Goal
from .player import Player
from .team import Team
from .game_info import GameInfo
from .frame_parser import parse_frames

logger = logging.getLogger(__name__)

DATETIME_FORMATS = [
    '%Y-%m-%d %H-%M-%S',
    '%Y-%m-%d:%H-%M'
]


class Game:

    def __init__(self):
        self.file_path = None
        self.replay = None
        self.properties = None
        self.replay_id = None
        self.map = None
        self.name = None
        self.id = None
        self.datetime = None
        self.replay_version = None

        self.replay_data = None

        self.teams: List[Team] = None  # Added in parse_all_data
        self.players: List[Player] = None
        self.goals: List[Goal] = None
        self.primary_player: dict = None
        self.all_data = None

        self.frames = None
        self.kickoff_frames = None
        self.ball = None
        self.ball_type = None
        self.demos = None
        self.parties = None
        self.dropshot = None

    def initialize(self, file_path='', loaded_json=None, parse_replay: bool = True, clean_player_names: bool = False):
        """
        Initializes the Game object by processing the replay's json file, which finds and copies all relevant data.

        :param file_path: The (string) path to the replay's json file.
        :param loaded_json: The replay's json file.
        :param parse_replay: Boolean - should the replay be parsed?
        :param clean_player_names: Boolean - should the player names be cleared?
        """

        self.file_path = file_path
        if loaded_json is None:
            with open(file_path, 'r') as f:
                self.replay = json.load(f)
        else:
            self.replay = loaded_json

        logger.debug('Loaded JSON')

        self.replay_data = self.replay['network_frames']['frames']

        # set properties
        self.properties = self.replay['properties']
        self.replay_id = self.properties['Id']
        self.map = self.properties.get('MapName', 'Unknown')
        self.name = self.properties.get('ReplayName', None)
        self.match_type = self.properties['MatchType']
        self.team_size = self.properties['TeamSize']

        if self.name is None:
            logger.warning('Replay name not found')
        self.id = self.properties["Id"]

        date_string = self.properties['Date']
        for date_format in DATETIME_FORMATS:
            try:
                self.datetime = datetime.strptime(date_string, date_format)
                break
            except ValueError:
                pass
        else:
            logger.error('Cannot parse date: ' + date_string)

        self.replay_version = self.properties.get('ReplayVersion', None)
        logger.info(f"version: {self.replay_version}, date: {self.datetime}")
        if self.replay_version is None:
            logger.warning('Replay version not found')

        self.players: List[Player] = self.create_players()
        self.goals: List[Goal] = self.get_goals()
        self.primary_player: dict = self.get_primary_player()

        if parse_replay:
            self.all_data = parse_frames(self)
            self.parse_all_data(self.all_data, clean_player_names)
            logger.info("Finished parsing %s" % self)

    def __repr__(self):
        team_0_name = self.teams[0].name
        team_1_name = self.teams[1].name
        team_0_score = self.teams[0].score
        team_1_score = self.teams[1].score
        return "%s: %s vs %s (%s:%s)" % (self.name, team_0_name, team_1_name, team_0_score, team_1_score)

    def create_players(self) -> List[Player]:
        players = []
        try:
            for player_stats in self.properties["PlayerStats"]:
                player = Player().parse_player_stats(player_stats)
                players.append(player)
        except KeyError:
            pass
        return players

    def get_primary_player(self):
        owner_name = self.properties.get('PlayerName')
        if owner_name is not None:
            for player in self.players:
                if player.name == owner_name:
                    return {'name': owner_name, 'id': player.online_id}
            return {'name': owner_name, 'id': None}

    def get_goals(self) -> List[Goal]:
        """
        Gets goals from replay_properties and creates respective Goal objects.

        :return: List[Goal]
        """

        if "Goals" not in self.properties:
            return []

        goals = self.properties["Goals"]

        logger.info('Found %s goals.' % len(goals))
        logger.debug('Goals: %s' % goals)

        goals_list = []
        for goal_dict in goals:
            goal = Goal(goal_dict, self)
            goals_list.append(goal)
        return goals_list


    def parse_all_data(self, all_data, clean_player_names: bool) -> None:
        """
        Finishes parsing after frame-parsing is done.
        E.g. Adds players not found in MatchStats metadata
        :param all_data: Dict returned by parse_replay
        :return:
        """
        # GAME INFO
        self.game_info = GameInfo().parse_game_info_actor(all_data['game_info_actor'],
                                                          all_data['soccar_game_event_actor'],
                                                          self.replay['objects'])

        # TEAMS
        self.teams = []
        for team_id, team_data in all_data['team_dicts'].items():
            team = Team().parse_team_data(team_data)
            self.teams.append(team)

        # PLAYERS
        player_actor_id_player_dict = {}  # player_actor_id: Player
        for _player_actor_id, _player_data in all_data['player_dicts'].items():
            if "TAGame.PRI_TA:MatchScore" not in _player_data:
                logger.warning(f"Player {_player_data['name']} as player has no MatchScore.")

            found_player = None
            for player in self.players:
                # if player leaves early, won't be created (as not found in metadata's player_stats)
                if _player_data['name'] == player.name:
                    found_player = player
                    player_actor_id_player_dict[_player_actor_id] = found_player
                    found_player.parse_actor_data(_player_data, self.replay['objects'])  # just add extra stuff
                    break
            if found_player is None:
                # player not in endgame stats, create new player
                try:
                    found_player = Player().create_from_actor_data(_player_data, self.teams, self.replay['objects'])
                except KeyError as e:
                    # KeyError: 'Engine.PlayerReplicationInfo:Team'
                    # in `team_actor_id = actor_data["Engine.PlayerReplicationInfo:Team"]`
                    # Player never actually joins the game.
                    if 'Engine.PlayerReplicationInfo:Team' not in _player_data:
                        logger.warning(f"Ignoring player: {_player_data['name']} as player has no team.")
                        continue
                    raise e
                self.players.append(found_player)
                player_actor_id_player_dict[_player_actor_id] = found_player

                # check if any goals are playerless and belong to this newly-created player
                for goal in self.goals:
                    if not goal.player and goal.player_name == found_player.name:
                        goal.player = found_player

            found_player.parse_data(all_data['player_ball_data'][_player_actor_id])
            # camera_settings might not exist (see 0AF8AC734890E6D3995B829E474F9924)
            found_player.get_camera_settings(all_data['cameras_data'].get(_player_actor_id, {}))
            found_player.get_data_from_car(all_data['car_dicts'].get(_player_actor_id, None))

            for team in self.teams:
                if found_player.is_orange == team.is_orange:
                    team.add_player(found_player)

            if clean_player_names:
                cleaned_player_name = re.sub(r'[^\x00-\x7f]', r'', found_player.name).strip()  # Support ASCII only
                if cleaned_player_name != found_player.name:
                    logger.warning(
                        f"Cleaned player name to ASCII-only. From: {found_player.name} to: {cleaned_player_name}")
                    found_player.name = cleaned_player_name

        # GOAL - add player if not found earlier (ie player just created)
        for goal in self.goals:
            if goal.player is None:
                goal.player = goal.get_player(self)

        # BALL
        self.ball = pd.DataFrame.from_dict(self.all_data['player_ball_data']['ball'], orient='index')

        # FRAMES
        self.frames = pd.DataFrame.from_dict(self.all_data['frames_data'], orient='index')

        # DEMOS
        self.demos = []
        demo_map = dict()
        for _demo_data in all_data['demos_data']:
            demo = {
                'frame_number': _demo_data['frame_number'],
                'attacker': player_actor_id_player_dict[_demo_data['attacker_player_id']],
                'victim': player_actor_id_player_dict[_demo_data['victim_player_id']],
                'attacker_vel': (_demo_data["attack_velocity"]["x"],
                                 _demo_data["attack_velocity"]["y"],
                                 _demo_data["attack_velocity"]["z"],),
                'victim_vel': (_demo_data["victim_velocity"]["x"],
                               _demo_data["victim_velocity"]["y"],
                               _demo_data["victim_velocity"]["z"],),
            }

            # Key created to prevent duplicate demo counts
            key = (int(_demo_data["attack_velocity"]["x"]) + int(_demo_data["attack_velocity"]["y"]) + int(_demo_data["attack_velocity"]["z"]) +
                   int(_demo_data["victim_velocity"]["x"]) + int(_demo_data["victim_velocity"]["y"]) + int(_demo_data["victim_velocity"]["z"]))
            Game.add_demo_to_map(key, demo, demo_map)

        self.demos = list(demo_map.values())

        # PARTIES
        self.parties = all_data['parties']

        # DROPSHOT EVENTS
        if 'dropshot_phase' in self.ball:
            self.ball['dropshot_phase'] = self.ball['dropshot_phase'].astype('int8')

        for column in self.frames.columns:
            if column.startswith('dropshot_tile'):
                self.frames[column] = self.frames[column].astype('int8')

        self.dropshot = {
            'damage_events': []
        }

        damage_events = all_data['dropshot']['damage_events']
        ball_events = all_data['dropshot']['ball_events']

        if len(damage_events) > 1:
            # sometimes damages can trickle over to the next frame, clean those up
            frames = list(damage_events.keys())

            i = 0
            while i < len(frames) - 1:
                if frames[i] + 1 == frames[i + 1] and damage_events[frames[i]][0] == damage_events[frames[i + 1]][0]:
                    ball_event = next(event for event in reversed(ball_events) if event['frame_number'] < frames[i])
                    ball_phase = ball_event['state']

                    first_frame = damage_events[frames[i]]
                    trickle_frame = frames[i + 1]

                    while trickle_frame in frames:
                        # check if the total damage of the two frames is not more than it could be
                        if (ball_phase == 1 and len(first_frame[1]) + len(damage_events[trickle_frame][1])) <= 7 or \
                                (ball_phase == 2 and len(first_frame[1]) + len(damage_events[trickle_frame][1]) <= 19):
                            first_frame[1].extend(damage_events[trickle_frame][1])
                            damage_events.pop(trickle_frame)
                            i += 1
                            trickle_frame += 1
                        else:
                            break
                i += 1

        for frame_number, damage in damage_events.items():
            self.dropshot['damage_events'].append({
                'frame_number': frame_number,
                'player': player_actor_id_player_dict[damage[0]],
                'tiles': damage[1]
            })

        damage_frames = set(damage_events.keys())
        self.dropshot['tile_frames'] = \
            {k: v for (k, v) in all_data['dropshot']['tile_frames'].items() if k in damage_frames}

        self.dropshot['ball_events'] = ball_events

        del self.replay_data
        del self.replay
        del self.all_data

    @staticmethod
    def add_demo_to_map(key, demo, demo_map):
        if key in demo_map:
            old_demo = demo_map[key]
            if demo['attacker'] == old_demo['attacker'] and demo['victim'] == old_demo['victim']:
                if demo['frame_number'] < old_demo['frame_number']:
                    demo_map[key] = demo
                return
            Game.add_demo_to_map(key + 1, demo, demo_map)
        else:
            demo_map[key] = demo



from carball.tests.utils import get_raw_replays, run_analysis_test_on_replay

from carball.analysis.analysis_manager import AnalysisManager


class Test_Demos:
    def test_calculate_demos_correctly(self, replay_cache):
        def test(analysis: AnalysisManager):
            proto_game = analysis.get_protobuf_data()
            bumps = proto_game.game_stats.bumps
            assert len(bumps) == 1

        run_analysis_test_on_replay(test, get_raw_replays()["1_DEMO"], cache=replay_cache)



import logging
import time
from typing import Dict, Callable

import pandas as pd
import json
import os

from google.protobuf.json_format import _Printer
from typing.io import IO

from .utils.json_encoder import CarballJsonEncoder

script_path = os.path.abspath(__file__)
with open(os.path.join(os.path.dirname(script_path), 'PROTOBUF_VERSION'), 'r') as f:
    PROTOBUF_VERSION = json.loads(f.read())

from ..analysis.cleaner.cleaner import clean_replay
from ..analysis.saltie_game.metadata.ApiGame import ApiGame
from ..analysis.saltie_game.metadata.ApiMutators import ApiMutators
from ..analysis.saltie_game.metadata.ApiPlayer import ApiPlayer
from ..analysis.saltie_game.metadata.ApiTeam import ApiTeam
from ..analysis.saltie_game.saltie_game import SaltieGame
from ..analysis.stats.stats_manager import StatsManager
from ..analysis.utils.pandas_manager import PandasManager
from ..analysis.utils.proto_manager import ProtobufManager
from ..generated.api import game_pb2
from ..generated.api.player_pb2 import Player
from ..json_parser.game import Game
from ..analysis.events.event_creator import EventsCreator

logger = logging.getLogger(__name__)


class AnalysisManager:
    """
    AnalysisManager class takes an initialized Game object and converts the data into a Protobuf and a DataFrame. Then,
    that data is used to perform full analysis on the replay.
    """

    id_creator = None
    timer = None

    def __init__(self, game: Game):
        self.game = game
        self.protobuf_game = game_pb2.Game()
        self.protobuf_game.version = PROTOBUF_VERSION
        self.id_creator = self._create_player_id_function(game)
        self.stats_manager = StatsManager()
        self.events_creator = EventsCreator(self.id_creator)
        self.should_store_frames = False
        self.df_bytes = None

    def create_analysis(self, calculate_intensive_events: bool = False, clean: bool = True):
        """
        Sets basic metadata, and decides whether analysis can be performed and then passes required parameters
        to perform_full_analysis(...); After, stores the DataFrame.

        :param calculate_intensive_events: Indicates if expensive calculations should run to include additional stats.
        :param clean: Indicates if useless/invalid data should be found and removed.
        """

        self._start_time()
        player_map = self._get_game_metadata(self.game, self.protobuf_game)
        self._log_time("Getting in-game frame-by-frame data...")
        data_frame = self._initialize_data_frame(self.game)
        self._log_time("Getting important frames (kickoff, first-touch)...")
        kickoff_frames, first_touch_frames = self._get_kickoff_frames(self.game, self.protobuf_game, data_frame)
        self._log_time("Setting game kickoff frames...")
        self.game.kickoff_frames = kickoff_frames

        if self._can_do_full_analysis(first_touch_frames):
            self._perform_full_analysis(self.game, self.protobuf_game, player_map,
                                        data_frame, kickoff_frames, first_touch_frames,
                                        calculate_intensive_events=calculate_intensive_events,
                                        clean=clean)
        else:
            self._log_time("Cannot perform analysis: invalid analysis.")
            self.protobuf_game.game_metadata.is_invalid_analysis = True

        # log before we add the dataframes
        # logger.debug(self.protobuf_game)

        self._store_frames(data_frame)

    def write_json_out_to_file(self, file: IO):
        """
        Writes the json data to the specified file, as text.

        NOTES:
            The data is written as text (i.e. string), and the buffer mode must be 'w'.
                E.g. open(file_name, 'w')

        :param file: The file object (or a buffer).
        """

        if 'b' in file.mode:
            raise IOError("Json files can not be binary use open(path,\"w\")")
        printer = _Printer()
        js = printer._MessageToJsonObject(self.protobuf_game)
        json.dump(js, file, indent=2, cls=CarballJsonEncoder)

    def write_proto_out_to_file(self, file: IO):
        """
        Writes the proto buffer data to the specified file, as bytes.

        NOTES:
            The data is written as bytes (i.e. in binary), and the buffer mode must be 'wb'.
                E.g. open(file_name, 'wb')
            The file will NOT be human-readable.

        :param file: The file object (or a buffer).
        """

        if 'b' not in file.mode:
            raise IOError("Proto files must be binary use open(path,\"wb\")")
        ProtobufManager.write_proto_out_to_file(file, self.protobuf_game)

    def write_pandas_out_to_file(self, file: IO):
        """
        Writes the pandas data to the specified file, as bytes.

        NOTES:
            The data is written as bytes (i.e. in binary), and the buffer mode must be 'wb'.
                E.g. open(file_name, 'wb')
            The file will NOT be human-readable.

        :param file: The file object (or a buffer).
        """

        if 'b' not in file.mode:
            raise IOError("Proto files must be binary use open(path,\"wb\")")
        if self.df_bytes is not None:
            file.write(self.df_bytes)
        elif not self.should_store_frames:
            logger.warning("pd DataFrames are not being stored anywhere")

    def get_protobuf_data(self) -> game_pb2.Game:
        """
        :return: The protobuf data created by the analysis

        USAGE: A Protocol Buffer contains in-game metadata (e.g. events, stats). Treat it as a usual Python object with
        fields that match the API.

        INFO: The Protocol Buffer is a collection of data organized in a format similar to json. All relevant .proto
        files found at https://github.com/SaltieRL/carball/tree/master/api.

        Google's developer guide to protocol buffers may be found at https://developers.google.com/protocol-buffers/docs/overview
        """
        return self.protobuf_game

    def get_json_data(self):
        """
        :return: The protobuf data created by the analysis as a json object.

        see get_protobuf_data for more details.
        The json fields are defined by https://github.com/SaltieRL/carball/tree/master/api
        """
        printer = _Printer()
        js = printer._MessageToJsonObject(self.protobuf_game)
        return js

    def get_data_frame(self) -> pd.DataFrame:
        """
        :return: The pandas.DataFrame object.

        USAGE: A DataFrame contains in-game frame-by-frame data.

        INFO: The DataFrame is a collection of data organized in a format similar to csv. The 'index' column of the
        DataFrame is the consecutive in-game frames, and all other column headings (150+) are tuples in the following
        format:
            (Object, Data), where the Object is either a player, the ball or the game.

        All column information (and keys) may be seen by calling data_frame.info(verbose=True)

        All further documentation about the DataFrame can be found at https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.html
        """
        return self.data_frame

    def _perform_full_analysis(self, game: Game, proto_game: game_pb2.Game, player_map: Dict[str, Player],
                               data_frame: pd.DataFrame, kickoff_frames: pd.DataFrame, first_touch_frames: pd.Series,
                               calculate_intensive_events: bool = False, clean: bool = True):

        """
        Sets some further data and cleans the replay;
        Then, performs the analysis.

        :param game: The game object (instance of Game). It contains the replay metadata and processed json data.
        :param proto_game: The game's protobuf (instance of game_pb2) (refer to the comment in get_protobuf_data() for more info).
        :param data_frame: The game's pandas.DataFrame object (refer to comment in get_data_frame() for more info).
        :param player_map: A map of player name to Player protobuf.
        :param kickoff_frames: Contains data about the kickoffs.
        :param first_touch_frames:  Contains data for frames where touches can actually occur.
        :param calculate_intensive_events: Indicates if expensive calculations should run to include additional stats.
        :param clean: Indicates if useless/invalid data should be found and removed.
        """

        self._get_game_time(proto_game, data_frame)
        if clean:
            clean_replay(game, data_frame, proto_game, player_map)
        self._log_time("Creating events...")
        self.events_creator.create_events(game, proto_game, player_map, data_frame, kickoff_frames, first_touch_frames,
                                          calculate_intensive_events=calculate_intensive_events)
        self._log_time("Getting stats...")
        self._get_stats(game, proto_game, player_map, data_frame)

    def _get_game_metadata(self, game: Game, proto_game: game_pb2.Game) -> Dict[str, Player]:
        """
        Processes protobuf data and sets the respective object fields to correct values.
        Maps the player's specific online ID (steam unique ID) to the player object.

        :param game: The game object (instance of Game). It contains the replay metadata and processed json data.
        :param proto_game: The game's protobuf (instance of game_pb2) (refer to the comment in get_protobuf_data() for more info).
        :return: A dictionary, with the player's online ID as the key, and the player object (instance of Player) as the value.
        """
        # Process the relevant protobuf data and pass it to the Game object (returned data is ignored).
        ApiGame.create_from_game(proto_game.game_metadata, game, self.id_creator)

        # Process the relevant protobuf data and pass it to the Game object's mutators (returned data is ignored).
        ApiMutators.create_from_game(proto_game.mutators, game, self.id_creator)

        # Process the relevant protobuf data and pass it to the Team objects (returned data is ignored).
        ApiTeam.create_teams_from_game(game, proto_game, self.id_creator)

        # Process the relevant protobuf data and add players to their respective parties.
        ApiGame.create_parties(proto_game.parties, game, self.id_creator)

        player_map = dict()
        for player in game.players:
            player_proto = proto_game.players.add()
            ApiPlayer.create_from_player(player_proto, player, self.id_creator)
            player_map[str(player.online_id)] = player_proto

        return player_map

    def _get_game_time(self, protobuf_game: game_pb2.Game, data_frame: pd.DataFrame):
        """
        Calculates the game length (total time the game lasted) and sets it to the relevant metadata length field.
        Calculates the total time a player has spent in the game and sets it to the relevant player field.

        :param proto_game: The game's protobuf (instance of game_pb2) (refer to the comment in get_protobuf_data() for more info).
        :param data_frame: The game's pandas.DataFrame object (refer to comment in get_data_frame() for more info).
        """

        protobuf_game.game_metadata.length = data_frame.game[data_frame.game.goal_number.notnull()].delta.sum()
        for player in protobuf_game.players:
            try:
                player.time_in_game = data_frame[
                    data_frame[player.name].pos_x.notnull() & data_frame.game.goal_number.notnull()].game.delta.sum()
                player.first_frame_in_game = data_frame[player.name].pos_x.first_valid_index()
            except:
                player.time_in_game = 0

        logger.info("Set each player's in-game times.")

    def _get_kickoff_frames(self, game: Game, proto_game: game_pb2.Game, data_frame: pd.DataFrame):
        """
        Firstly, fetches kickoff-related data from SaltieGame.
        Secondly, checks for edge-cases and corrects errors.

        NOTE: kickoff_frames is an array of all in-game frames at each kickoff beginning.
        NOTE: first_touch_frames is an array of all in-game frames for each 'First Touch' at kickoff.

        :param game: The game object (instance of Game). It contains the replay metadata and processed json data.
        :param proto_game: The game's protobuf (instance of game_pb2) (refer to the comment in get_protobuf_data() for more info).
        :param data_frame: The game's pandas.DataFrame object (refer to comment in get_data_frame() for more info).
        :return: See notes above.
        """

        kickoff_frames = SaltieGame.get_kickoff_frames(game)
        first_touch_frames = SaltieGame.get_first_touch_frames(game)

        if len(kickoff_frames) > len(first_touch_frames):
            # happens when the game ends before anyone touches the ball at kickoff
            kickoff_frames = kickoff_frames[:len(first_touch_frames)]

        for goal_number, goal in enumerate(game.goals):
            data_frame.loc[kickoff_frames[goal_number]: goal.frame_number, ('game', 'goal_number')] = goal_number

        # Set goal_number of frames that are post last kickoff to -1 (ie non None)
        if len(kickoff_frames) > len(proto_game.game_metadata.goals):
            data_frame.loc[kickoff_frames[-1]:, ('game', 'goal_number')] = -1

        for index in range(min(len(kickoff_frames), len(first_touch_frames))):
            kickoff = proto_game.game_stats.kickoffs.add()
            kickoff.start_frame_number = kickoff_frames[index]
            kickoff.end_frame_number = first_touch_frames[index]

        return kickoff_frames, first_touch_frames

    def _get_stats(self, game: Game, proto_game: game_pb2.Game, player_map: Dict[str, Player],
                   data_frame: pd.DataFrame):
        """
        For each in-game frame after a goal has happened, calculate in-game stats.
        (i.e. player, team, general-game and hit stats)

        :param game: The game object (instance of Game). It contains the replay metadata and processed json data.
        :param proto_game: The game's protobuf (instance of game_pb2) (refer to the comment in get_protobuf_data() for more info).
        :param data_frame: The game's pandas.DataFrame object (refer to comment in get_data_frame() for more info).
        :param player_map: The dictionary with all player IDs matched to the player objects.
        """

        goal_frames = data_frame.game.goal_number.notnull()
        self.stats_manager.get_stats(game, proto_game, player_map, data_frame[goal_frames])

    def _store_frames(self, data_frame: pd.DataFrame):
        self.data_frame = data_frame
        self.df_bytes = PandasManager.safe_write_pandas_to_memory(data_frame)

    def _initialize_data_frame(self, game: Game):
        data_frame = SaltieGame.create_data_df(game)

        logger.info("Assigned goal_number in .data_frame")
        return data_frame

    def _create_player_id_function(self, game: Game) -> Callable:
        name_map = {player.name: player.online_id for player in game.players}

        def create_name(proto_player_id, name):
            proto_player_id.id = str(name_map[name])

        return create_name

    def _can_do_full_analysis(self, first_touch_frames) -> bool:
        """
        Check whether or not the replay satisfies the requirements for a full analysis.
        This includes checking:
            if at least 1 team is present;
            if the ball was touched at least once;
            if both teams have an equal amount of players.

        In some cases, the check for equal team sizes must be ignored due to spectators joining the match, for example.
        Therefore, this check is ignored by default.

        :param first_touch_frames: An array of all in-game frames for each 'First Touch' at kickoff.
        :return: Bool - true if criteria above are satisfied.
        """

        team_sizes = []
        for team in self.game.teams:
            team_sizes.append(len(team.players))

        if len(team_sizes) == 0:
            logger.warning("Not doing full analysis. No teams found")
            return False

        if len(first_touch_frames) == 0:
            logger.warning("Not doing full analysis. No one touched the ball")
            return False

        # if any((team_size != team_sizes[0]) for team_size in team_sizes):
        #     logger.warning("Not doing full analysis. Not all team sizes are equal")
        #     return False

        return True

    def _start_time(self):
        self.timer = time.time()
        logger.info("starting timer")

    def _log_time(self, message=""):
        end = time.time()
        logger.info("Time taken for %s is %s milliseconds", message, (end - self.timer) * 1000)
        self.timer = end



from typing import Callable

from ....json_parser.game import Game


class ApiDemo:
    @staticmethod
    def create_demos_from_game(game: Game, proto_demo_list, id_creator: Callable):
        for demo in game.demos:
            proto_demo = proto_demo_list.add()
            proto_demo.frame_number = demo['frame_number']
            id_creator(proto_demo.attacker_id, demo['attacker'].name)
            id_creator(proto_demo.victim_id, demo['victim'].name)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: api/stats/goal.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='api/stats/goal.proto',
  package='api.stats',
  serialized_pb=_b('\n\x14\x61pi/stats/goal.proto\x12\tapi.stats\"\x1f\n\tGoalStats\x12\x12\n\ngoal_index\x18\x01 \x01(\x05')
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_GOALSTATS = _descriptor.Descriptor(
  name='GoalStats',
  full_name='api.stats.GoalStats',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='goal_index', full_name='api.stats.GoalStats.goal_index', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=35,
  serialized_end=66,
)

DESCRIPTOR.message_types_by_name['GoalStats'] = _GOALSTATS

GoalStats = _reflection.GeneratedProtocolMessageType('GoalStats', (_message.Message,), dict(
  DESCRIPTOR = _GOALSTATS,
  __module__ = 'api.stats.goal_pb2'
  # @@protoc_insertion_point(class_scope:api.stats.GoalStats)
  ))
_sym_db.RegisterMessage(GoalStats)


# @@protoc_insertion_point(module_scope)



try:
    from carball.decompile_replays import decompile_replay
    from carball.decompile_replays import analyze_replay_file
except ModuleNotFoundError as e:
    print("Not importing functions due to missing packages:", e)


from .base import *
from carball.generated.api.game_pb2 import mutators_pb2 as mutators
from carball.json_parser.actor_parsing import BallActor

BALL_TYPES = {
    'Archetypes.Ball.Ball_Default': mutators.DEFAULT,
    'Archetypes.Ball.Ball_Basketball': mutators.BASKETBALL,
    'Archetypes.Ball.Ball_Puck': mutators.PUCK,
    'Archetypes.Ball.CubeBall': mutators.CUBEBALL,
    'Archetypes.Ball.Ball_Breakout': mutators.BREAKOUT
}


class BallHandler(BaseActorHandler):

    @classmethod
    def can_handle(cls, actor: dict) -> bool:
        return actor['TypeName'].startswith('Archetypes.Ball.')

    def update(self, actor: dict, frame_number: int, time: float, delta: float) -> None:
        if actor.get('TAGame.RBActor_TA:bIgnoreSyncing', False):
            return

        if self.parser.game.ball_type is None:
            self.parser.game.ball_type = BALL_TYPES.get(actor['TypeName'], mutators.DEFAULT)

        ball_data = BallActor.get_data_dict(actor)
        self.parser.ball_data[frame_number] = ball_data

        if self.parser.game.ball_type == mutators.BREAKOUT:
            damage_index = actor.get('TAGame.Ball_Breakout_TA:DamageIndex', 0)
            if damage_index > self.parser.dropshot['ball_state']:
                team = actor.get('TAGame.Ball_Breakout_TA:LastTeamTouch')
                self.parser.dropshot['ball_events'].append({
                    'state': damage_index,
                    'frame_number': frame_number,
                    'team': team
                })
            self.parser.dropshot['ball_state'] = damage_index
            self.parser.ball_data[frame_number]['dropshot_phase'] = damage_index



import logging

import numpy as np

from .car import read_xl

logger = logging.getLogger(__name__)


class Hitbox:
    xl = read_xl()

    def __init__(self, car_item_id: int):
        column_names = ['Length', 'Width', 'Height', 'Offset', 'Elevation']
        try:
            car_hitbox = self.xl.loc[car_item_id, column_names].values
        except KeyError:
            logger.debug("Cannot find car body id: %s. Falling back onto Octane." % car_item_id)
            car_hitbox = self.xl.loc[23, column_names].values

        self.car_length, self.car_width, self.car_height, self.car_offset, self.car_elevation = car_hitbox
        self.car_x_lims = (-self.car_length / 2 + self.car_offset,
                           self.car_length / 2 + self.car_offset)
        self.car_y_lims = (-self.car_width / 2,
                           self.car_width / 2)
        self.car_z_lims = (-self.car_height / 2 + self.car_elevation,
                           self.car_height / 2 + self.car_elevation)

    def get_collision_distance(self, ball_displacement) -> np.float64:
        if np.isnan(ball_displacement).any():
            return np.nan
        pos_x, pos_y, pos_z = ball_displacement
        # x axis
        if pos_x < self.car_x_lims[0]:
            x_dist = abs(self.car_x_lims[0] - pos_x)
        elif pos_x > self.car_x_lims[1]:
            x_dist = abs(self.car_x_lims[1] - pos_x)
        else:
            x_dist = 0
        # y axis
        if pos_y < self.car_y_lims[0]:
            y_dist = abs(self.car_y_lims[0] - pos_y)
        elif pos_y > self.car_y_lims[1]:
            y_dist = abs(self.car_y_lims[1] - pos_y)
        else:
            y_dist = 0
        # z axis
        if pos_z < self.car_z_lims[0]:
            z_dist = abs(self.car_z_lims[0] - pos_z)
        elif pos_z > self.car_z_lims[1]:
            z_dist = abs(self.car_z_lims[1] - pos_z)
        else:
            z_dist = 0

        return (x_dist ** 2 + y_dist ** 2 + z_dist ** 2) ** 0.5



from carball.analysis.analysis_manager import AnalysisManager

from carball.json_parser.game import Game
from carball.tests.utils import run_tests_on_list, run_analysis_test_on_replay, get_raw_replays

from carball import decompile_replays


class Test_OverallFunctionality():
    def test_replay_attrs(self):

        def test(replay):
            json_object = decompile_replays.decompile_replay(replay)
            game = Game()
            game.initialize(loaded_json=json_object)
            assert game.game_info.server_name is not None
            assert game.map is not None
            assert game.game_info.match_guid is not None

        run_tests_on_list(test, get_raw_replays()["0_JUMPS"])

    def test_full_replays(self, replay_cache):

        def test(analysis: AnalysisManager):
            assert (analysis.get_protobuf_data() is not None)
            assert(not analysis.get_protobuf_data().game_metadata.is_invalid_analysis)
            for player in analysis.get_protobuf_data().players:
                ratio = (player.stats.positional_tendencies.time_in_front_ball +
                         player.stats.positional_tendencies.time_behind_ball) / player.time_in_game
                assert ratio > 0.99
                # local.assertGreater(player.stats.positional_tendencies.time_in_front_ball, 0)
                # local.assertGreater(player.stats.positional_tendencies.time_behind_ball, 0)
                assert (player.time_in_game > 0)
                assert(player.stats.speed.time_at_slow_speed > 0)
                assert(player.stats.boost.average_boost_level > 0)
                assert(player.stats.boost.wasted_collection > -1)
            json = analysis.get_json_data()
            assert len(json['players']) > 0

        run_analysis_test_on_replay(test, cache=replay_cache)

    def test_unicode_error(self, replay_cache):

        def test(analysis: AnalysisManager):
            assert (analysis.get_protobuf_data() is not None)
            assert (not analysis.get_protobuf_data().game_metadata.is_invalid_analysis)
            for player in analysis.get_protobuf_data().players:
                ratio = (player.stats.positional_tendencies.time_in_front_ball +
                         player.stats.positional_tendencies.time_behind_ball) / player.time_in_game
                assert (ratio > 0.99)
                assert (player.stats.positional_tendencies.time_in_front_ball > 0)
                assert (player.stats.positional_tendencies.time_behind_ball > 0)
                assert (player.time_in_game > 0)
                assert (player.stats.speed.time_at_slow_speed > 0)
                assert (player.stats.boost.average_boost_level > 0)

        run_analysis_test_on_replay(test, get_raw_replays()['UNICODE_ERROR'], cache=replay_cache)



import logging
from typing import Dict

import pandas as pd

from carball.analysis.stats.stats import BaseStat
from carball.generated.api import game_pb2
from carball.generated.api.player_pb2 import Player
from carball.generated.api.stats.team_stats_pb2 import TeamStats
from carball.generated.api.stats.rumble_pb2 import PowerUp
from carball.json_parser.game import Game
from carball.analysis.stats.rumble.rumble import is_rumble_enabled

log = logging.getLogger(__name__)


class PreRumbleGoals(BaseStat):

    def calculate_team_stat(self, team_stat_list: Dict[int, TeamStats], game: Game, proto_game: game_pb2.Game,
                            player_map: Dict[str, Player], data_frame: pd.DataFrame):
        if not is_rumble_enabled(game):
            return

        pre_power_up_goals = [0, 0]

        item_get_frames = sorted(set(map(lambda event: event.frame_number_get, proto_game.game_stats.rumble_items)))

        for goal in proto_game.game_metadata.goals:
            # kick off frame before the goal
            try:
                kickoff_frame = next(frame for frame in reversed(game.kickoff_frames) if frame < goal.frame_number)
            except StopIteration:
                # there was no kickoff before game, the replay started mid game
                log.warning(f'There was a goal at frame {goal.frame_number} before the replay started. '
                            f'Cannot tell if it was pre rumble items.')
                continue

            # first item get frame after kick off
            next_get_frame = next((frame for frame in item_get_frames if frame > kickoff_frame), -1)

            if next_get_frame > goal.frame_number or next_get_frame == -1:
                # goal before rumble items
                if goal.player_id.id in player_map:
                    pre_power_up_goals[player_map[goal.player_id.id].is_orange] += 1
                else:
                    log.error(f'Could not determine which team the goal at {goal.frame_number} belongs to, '
                              f'missing player {goal.player_id.id}')
                goal.extra_mode_info.pre_items = True
            else:
                goal.extra_mode_info.pre_items = False

        team_stat_list[0].rumble_stats.pre_item_goals = pre_power_up_goals[0]
        team_stat_list[1].rumble_stats.pre_item_goals = pre_power_up_goals[1]


class ItemGoals(BaseStat):

    def calculate_stat(self, proto_stat, game: Game, proto_game: game_pb2.Game, player_map: Dict[str, Player],
                       data_frame: pd.DataFrame):
        if not is_rumble_enabled(game):
            return

        game_df = data_frame['game']

        for goal in proto_game.game_metadata.goals:
            # time of the goal
            goal_time = game_df.loc[goal.frame_number]['time']

            # get the frame number 3 seconds before the goal
            start_frame = game_df.loc[game_df['time'] >= goal_time - 3].index[0]

            player_df = data_frame[player_map[goal.player_id.id].name]
            player_df = player_df.loc[start_frame:goal.frame_number]

            if player_df['power_up_active'].any():
                goal.extra_mode_info.scored_with_item = True
                goal.extra_mode_info.used_item = PowerUp.Value(player_df.loc[player_df['power_up_active'] == True]
                                                           .iloc[0]['power_up'].upper())
            else:
                goal.extra_mode_info.scored_with_item = False



COMPONENT_REPLICATED_ACTIVE_KEY = 'TAGame.CarComponent_TA:ReplicatedActive'
REPLICATED_RB_STATE_KEY = 'TAGame.RBActor_TA:ReplicatedRBState'
COMPONENT_ACTIVE_KEY = 'TAGame.CarComponent_TA:Active'


class BaseActorHandler(object):
    """
    Base class for actor handlers.

    Attributes:
        type_name (str): If can_handle is not overridden, it will check the type_name attribute.
    """
    type_name: str = None

    @classmethod
    def can_handle(cls, actor: dict) -> bool:
        """
        :param actor: the actor from the json frame
        :return: True if the the class can handle the actor
        """
        return cls.type_name is not None and cls.type_name == actor['TypeName']

    def __init__(self, parser: object):
        """
        :param parser: The FrameParser object
        """
        self.parser = parser

    def update(self, actor: dict, frame_number: int, time: float, delta: float) -> None:
        """
        Called for each frame the actor of this handler is alive for.

        :param actor: the actor of this handler
        :param frame_number: the number of the frame
        :param time: frame time
        :param delta: frame time delta
        """
        pass



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: api/stats/player_stats.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ...api.stats import stats_pb2
from ...api.stats import events_pb2
from ...api.stats import per_possession_stats_pb2
from ...api.stats import rumble_pb2
from ...api.stats import dropshot_pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='api/stats/player_stats.proto',
  package='api.stats',
  serialized_pb=_b('\n\x1c\x61pi/stats/player_stats.proto\x12\tapi.stats\x1a\x15\x61pi/stats/stats.proto\x1a\x16\x61pi/stats/events.proto\x1a$api/stats/per_possession_stats.proto\x1a\x16\x61pi/stats/rumble.proto\x1a\x18\x61pi/stats/dropshot.proto\"\xab\x02\n\x05\x42oost\x12\x13\n\x0b\x62oost_usage\x18\x01 \x01(\x01\x12\x18\n\x10num_small_boosts\x18\x02 \x01(\x05\x12\x18\n\x10num_large_boosts\x18\x03 \x01(\x05\x12\x19\n\x11wasted_collection\x18\x04 \x01(\x02\x12\x14\n\x0cwasted_usage\x18\x05 \x01(\x02\x12\x17\n\x0ftime_full_boost\x18\x06 \x01(\x02\x12\x16\n\x0etime_low_boost\x18\x07 \x01(\x02\x12\x15\n\rtime_no_boost\x18\x08 \x01(\x02\x12\x19\n\x11num_stolen_boosts\x18\t \x01(\x05\x12\x1b\n\x13\x61verage_boost_level\x18\n \x01(\x02\x12\x12\n\nwasted_big\x18\x0b \x01(\x02\x12\x14\n\x0cwasted_small\x18\x0c \x01(\x02\"\xe7\x01\n\x08\x44istance\x12\x18\n\x10\x62\x61ll_hit_forward\x18\x01 \x01(\x01\x12\x19\n\x11\x62\x61ll_hit_backward\x18\x02 \x01(\x01\x12\x1c\n\x14time_closest_to_ball\x18\x03 \x01(\x02\x12\x1f\n\x17time_furthest_from_ball\x18\x04 \x01(\x02\x12\x1a\n\x12time_close_to_ball\x18\x05 \x01(\x02\x12#\n\x1btime_closest_to_team_center\x18\x06 \x01(\x02\x12&\n\x1etime_furthest_from_team_center\x18\x07 \x01(\x02\"\xc1\x01\n\x13RelativePositioning\x12\'\n\x1ftime_in_front_of_center_of_mass\x18\x01 \x01(\x02\x12\"\n\x1atime_behind_center_of_mass\x18\x02 \x01(\x02\x12 \n\x18time_most_forward_player\x18\x03 \x01(\x02\x12\x1d\n\x15time_most_back_player\x18\x04 \x01(\x02\x12\x1c\n\x14time_between_players\x18\x05 \x01(\x02\"\xd3\x05\n\x0bPlayerStats\x12\x1f\n\x05\x62oost\x18\x01 \x01(\x0b\x32\x10.api.stats.Boost\x12%\n\x08\x64istance\x18\x02 \x01(\x0b\x32\x13.api.stats.Distance\x12)\n\npossession\x18\x03 \x01(\x0b\x32\x15.api.stats.Possession\x12>\n\x15positional_tendencies\x18\x04 \x01(\x0b\x32\x1f.api.stats.PositionalTendencies\x12%\n\x08\x61verages\x18\x05 \x01(\x0b\x32\x13.api.stats.Averages\x12(\n\nhit_counts\x18\x06 \x01(\x0b\x32\x14.api.stats.HitCounts\x12/\n\x0e\x63\x61mera_changes\x18\x07 \x03(\x0b\x32\x17.api.stats.CameraChange\x12)\n\ncontroller\x18\x08 \x01(\x0b\x32\x15.api.stats.Controller\x12\x1f\n\x05speed\x18\t \x01(\x0b\x32\x10.api.stats.Speed\x12<\n\x14relative_positioning\x18\n \x01(\x0b\x32\x1e.api.stats.RelativePositioning\x12;\n\x14per_possession_stats\x18\x0b \x01(\x0b\x32\x1d.api.stats.PerPossessionStats\x12,\n\x0crumble_stats\x18\x0c \x01(\x0b\x32\x16.api.stats.RumbleStats\x12.\n\x0c\x62\x61ll_carries\x18\r \x01(\x0b\x32\x18.api.stats.CarryDribbles\x12\x38\n\rkickoff_stats\x18\x0e \x01(\x0b\x32!.api.stats.CumulativeKickoffStats\x12\x30\n\x0e\x64ropshot_stats\x18\x0f \x01(\x0b\x32\x18.api.stats.DropshotStats\"s\n\nController\x12\x13\n\x0bis_keyboard\x18\x01 \x01(\x08\x12\'\n\x1f\x61nalogue_steering_input_percent\x18\x02 \x01(\x02\x12\'\n\x1f\x61nalogue_throttle_input_percent\x18\x03 \x01(\x02\"\x90\x02\n\rCarryDribbles\x12\x15\n\rtotal_carries\x18\x01 \x01(\x05\x12\x14\n\x0ctotal_flicks\x18\x02 \x01(\x05\x12\x15\n\rlongest_carry\x18\x03 \x01(\x02\x12\x16\n\x0e\x66urthest_carry\x18\x04 \x01(\x02\x12\x18\n\x10total_carry_time\x18\x05 \x01(\x02\x12\x1a\n\x12\x61verage_carry_time\x18\x06 \x01(\x02\x12\x1b\n\x13\x66\x61stest_carry_speed\x18\x07 \x01(\x02\x12\x1c\n\x14total_carry_distance\x18\x08 \x01(\x02\x12\x32\n\x0b\x63\x61rry_stats\x18\t \x01(\x0b\x32\x1d.api.stats.DetailedCarryStats\"\xe6\x01\n\x16\x43umulativeKickoffStats\x12\x16\n\x0etotal_kickoffs\x18\x01 \x01(\x05\x12\x16\n\x0enum_time_boost\x18\x02 \x01(\x05\x12\x16\n\x0enum_time_cheat\x18\x03 \x01(\x05\x12\x17\n\x0fnum_time_defend\x18\x04 \x01(\x05\x12\x1b\n\x13num_time_go_to_ball\x18\x05 \x01(\x05\x12\x14\n\x0cnum_time_afk\x18\x06 \x01(\x05\x12\x1c\n\x14num_time_first_touch\x18\x07 \x01(\x05\x12\x1a\n\x12\x61verage_boost_used\x18\x08 \x01(\x02')
  ,
  dependencies=[stats_pb2.DESCRIPTOR,events_pb2.DESCRIPTOR,per_possession_stats_pb2.DESCRIPTOR,rumble_pb2.DESCRIPTOR,dropshot_pb2.DESCRIPTOR,])
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_BOOST = _descriptor.Descriptor(
  name='Boost',
  full_name='api.stats.Boost',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='boost_usage', full_name='api.stats.Boost.boost_usage', index=0,
      number=1, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='num_small_boosts', full_name='api.stats.Boost.num_small_boosts', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='num_large_boosts', full_name='api.stats.Boost.num_large_boosts', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='wasted_collection', full_name='api.stats.Boost.wasted_collection', index=3,
      number=4, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='wasted_usage', full_name='api.stats.Boost.wasted_usage', index=4,
      number=5, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='time_full_boost', full_name='api.stats.Boost.time_full_boost', index=5,
      number=6, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='time_low_boost', full_name='api.stats.Boost.time_low_boost', index=6,
      number=7, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='time_no_boost', full_name='api.stats.Boost.time_no_boost', index=7,
      number=8, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='num_stolen_boosts', full_name='api.stats.Boost.num_stolen_boosts', index=8,
      number=9, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='average_boost_level', full_name='api.stats.Boost.average_boost_level', index=9,
      number=10, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='wasted_big', full_name='api.stats.Boost.wasted_big', index=10,
      number=11, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='wasted_small', full_name='api.stats.Boost.wasted_small', index=11,
      number=12, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=179,
  serialized_end=478,
)


_DISTANCE = _descriptor.Descriptor(
  name='Distance',
  full_name='api.stats.Distance',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='ball_hit_forward', full_name='api.stats.Distance.ball_hit_forward', index=0,
      number=1, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='ball_hit_backward', full_name='api.stats.Distance.ball_hit_backward', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='time_closest_to_ball', full_name='api.stats.Distance.time_closest_to_ball', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='time_furthest_from_ball', full_name='api.stats.Distance.time_furthest_from_ball', index=3,
      number=4, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='time_close_to_ball', full_name='api.stats.Distance.time_close_to_ball', index=4,
      number=5, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='time_closest_to_team_center', full_name='api.stats.Distance.time_closest_to_team_center', index=5,
      number=6, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='time_furthest_from_team_center', full_name='api.stats.Distance.time_furthest_from_team_center', index=6,
      number=7, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=481,
  serialized_end=712,
)


_RELATIVEPOSITIONING = _descriptor.Descriptor(
  name='RelativePositioning',
  full_name='api.stats.RelativePositioning',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='time_in_front_of_center_of_mass', full_name='api.stats.RelativePositioning.time_in_front_of_center_of_mass', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='time_behind_center_of_mass', full_name='api.stats.RelativePositioning.time_behind_center_of_mass', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='time_most_forward_player', full_name='api.stats.RelativePositioning.time_most_forward_player', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='time_most_back_player', full_name='api.stats.RelativePositioning.time_most_back_player', index=3,
      number=4, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='time_between_players', full_name='api.stats.RelativePositioning.time_between_players', index=4,
      number=5, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=715,
  serialized_end=908,
)


_PLAYERSTATS = _descriptor.Descriptor(
  name='PlayerStats',
  full_name='api.stats.PlayerStats',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='boost', full_name='api.stats.PlayerStats.boost', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='distance', full_name='api.stats.PlayerStats.distance', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='possession', full_name='api.stats.PlayerStats.possession', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='positional_tendencies', full_name='api.stats.PlayerStats.positional_tendencies', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='averages', full_name='api.stats.PlayerStats.averages', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='hit_counts', full_name='api.stats.PlayerStats.hit_counts', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='camera_changes', full_name='api.stats.PlayerStats.camera_changes', index=6,
      number=7, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='controller', full_name='api.stats.PlayerStats.controller', index=7,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='speed', full_name='api.stats.PlayerStats.speed', index=8,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='relative_positioning', full_name='api.stats.PlayerStats.relative_positioning', index=9,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='per_possession_stats', full_name='api.stats.PlayerStats.per_possession_stats', index=10,
      number=11, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='rumble_stats', full_name='api.stats.PlayerStats.rumble_stats', index=11,
      number=12, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='ball_carries', full_name='api.stats.PlayerStats.ball_carries', index=12,
      number=13, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='kickoff_stats', full_name='api.stats.PlayerStats.kickoff_stats', index=13,
      number=14, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='dropshot_stats', full_name='api.stats.PlayerStats.dropshot_stats', index=14,
      number=15, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=911,
  serialized_end=1634,
)


_CONTROLLER = _descriptor.Descriptor(
  name='Controller',
  full_name='api.stats.Controller',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='is_keyboard', full_name='api.stats.Controller.is_keyboard', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='analogue_steering_input_percent', full_name='api.stats.Controller.analogue_steering_input_percent', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='analogue_throttle_input_percent', full_name='api.stats.Controller.analogue_throttle_input_percent', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1636,
  serialized_end=1751,
)


_CARRYDRIBBLES = _descriptor.Descriptor(
  name='CarryDribbles',
  full_name='api.stats.CarryDribbles',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='total_carries', full_name='api.stats.CarryDribbles.total_carries', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='total_flicks', full_name='api.stats.CarryDribbles.total_flicks', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='longest_carry', full_name='api.stats.CarryDribbles.longest_carry', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='furthest_carry', full_name='api.stats.CarryDribbles.furthest_carry', index=3,
      number=4, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='total_carry_time', full_name='api.stats.CarryDribbles.total_carry_time', index=4,
      number=5, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='average_carry_time', full_name='api.stats.CarryDribbles.average_carry_time', index=5,
      number=6, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='fastest_carry_speed', full_name='api.stats.CarryDribbles.fastest_carry_speed', index=6,
      number=7, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='total_carry_distance', full_name='api.stats.CarryDribbles.total_carry_distance', index=7,
      number=8, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='carry_stats', full_name='api.stats.CarryDribbles.carry_stats', index=8,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1754,
  serialized_end=2026,
)


_CUMULATIVEKICKOFFSTATS = _descriptor.Descriptor(
  name='CumulativeKickoffStats',
  full_name='api.stats.CumulativeKickoffStats',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='total_kickoffs', full_name='api.stats.CumulativeKickoffStats.total_kickoffs', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='num_time_boost', full_name='api.stats.CumulativeKickoffStats.num_time_boost', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='num_time_cheat', full_name='api.stats.CumulativeKickoffStats.num_time_cheat', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='num_time_defend', full_name='api.stats.CumulativeKickoffStats.num_time_defend', index=3,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='num_time_go_to_ball', full_name='api.stats.CumulativeKickoffStats.num_time_go_to_ball', index=4,
      number=5, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='num_time_afk', full_name='api.stats.CumulativeKickoffStats.num_time_afk', index=5,
      number=6, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='num_time_first_touch', full_name='api.stats.CumulativeKickoffStats.num_time_first_touch', index=6,
      number=7, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='average_boost_used', full_name='api.stats.CumulativeKickoffStats.average_boost_used', index=7,
      number=8, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2029,
  serialized_end=2259,
)

_PLAYERSTATS.fields_by_name['boost'].message_type = _BOOST
_PLAYERSTATS.fields_by_name['distance'].message_type = _DISTANCE
_PLAYERSTATS.fields_by_name['possession'].message_type = stats_pb2._POSSESSION
_PLAYERSTATS.fields_by_name['positional_tendencies'].message_type = stats_pb2._POSITIONALTENDENCIES
_PLAYERSTATS.fields_by_name['averages'].message_type = stats_pb2._AVERAGES
_PLAYERSTATS.fields_by_name['hit_counts'].message_type = stats_pb2._HITCOUNTS
_PLAYERSTATS.fields_by_name['camera_changes'].message_type = events_pb2._CAMERACHANGE
_PLAYERSTATS.fields_by_name['controller'].message_type = _CONTROLLER
_PLAYERSTATS.fields_by_name['speed'].message_type = stats_pb2._SPEED
_PLAYERSTATS.fields_by_name['relative_positioning'].message_type = _RELATIVEPOSITIONING
_PLAYERSTATS.fields_by_name['per_possession_stats'].message_type = per_possession_stats_pb2._PERPOSSESSIONSTATS
_PLAYERSTATS.fields_by_name['rumble_stats'].message_type = rumble_pb2._RUMBLESTATS
_PLAYERSTATS.fields_by_name['ball_carries'].message_type = _CARRYDRIBBLES
_PLAYERSTATS.fields_by_name['kickoff_stats'].message_type = _CUMULATIVEKICKOFFSTATS
_PLAYERSTATS.fields_by_name['dropshot_stats'].message_type = dropshot_pb2._DROPSHOTSTATS
_CARRYDRIBBLES.fields_by_name['carry_stats'].message_type = stats_pb2._DETAILEDCARRYSTATS
DESCRIPTOR.message_types_by_name['Boost'] = _BOOST
DESCRIPTOR.message_types_by_name['Distance'] = _DISTANCE
DESCRIPTOR.message_types_by_name['RelativePositioning'] = _RELATIVEPOSITIONING
DESCRIPTOR.message_types_by_name['PlayerStats'] = _PLAYERSTATS
DESCRIPTOR.message_types_by_name['Controller'] = _CONTROLLER
DESCRIPTOR.message_types_by_name['CarryDribbles'] = _CARRYDRIBBLES
DESCRIPTOR.message_types_by_name['CumulativeKickoffStats'] = _CUMULATIVEKICKOFFSTATS

Boost = _reflection.GeneratedProtocolMessageType('Boost', (_message.Message,), dict(
  DESCRIPTOR = _BOOST,
  __module__ = 'api.stats.player_stats_pb2'
  # @@protoc_insertion_point(class_scope:api.stats.Boost)
  ))
_sym_db.RegisterMessage(Boost)

Distance = _reflection.GeneratedProtocolMessageType('Distance', (_message.Message,), dict(
  DESCRIPTOR = _DISTANCE,
  __module__ = 'api.stats.player_stats_pb2'
  # @@protoc_insertion_point(class_scope:api.stats.Distance)
  ))
_sym_db.RegisterMessage(Distance)

RelativePositioning = _reflection.GeneratedProtocolMessageType('RelativePositioning', (_message.Message,), dict(
  DESCRIPTOR = _RELATIVEPOSITIONING,
  __module__ = 'api.stats.player_stats_pb2'
  # @@protoc_insertion_point(class_scope:api.stats.RelativePositioning)
  ))
_sym_db.RegisterMessage(RelativePositioning)

PlayerStats = _reflection.GeneratedProtocolMessageType('PlayerStats', (_message.Message,), dict(
  DESCRIPTOR = _PLAYERSTATS,
  __module__ = 'api.stats.player_stats_pb2'
  # @@protoc_insertion_point(class_scope:api.stats.PlayerStats)
  ))
_sym_db.RegisterMessage(PlayerStats)

Controller = _reflection.GeneratedProtocolMessageType('Controller', (_message.Message,), dict(
  DESCRIPTOR = _CONTROLLER,
  __module__ = 'api.stats.player_stats_pb2'
  # @@protoc_insertion_point(class_scope:api.stats.Controller)
  ))
_sym_db.RegisterMessage(Controller)

CarryDribbles = _reflection.GeneratedProtocolMessageType('CarryDribbles', (_message.Message,), dict(
  DESCRIPTOR = _CARRYDRIBBLES,
  __module__ = 'api.stats.player_stats_pb2'
  # @@protoc_insertion_point(class_scope:api.stats.CarryDribbles)
  ))
_sym_db.RegisterMessage(CarryDribbles)

CumulativeKickoffStats = _reflection.GeneratedProtocolMessageType('CumulativeKickoffStats', (_message.Message,), dict(
  DESCRIPTOR = _CUMULATIVEKICKOFFSTATS,
  __module__ = 'api.stats.player_stats_pb2'
  # @@protoc_insertion_point(class_scope:api.stats.CumulativeKickoffStats)
  ))
_sym_db.RegisterMessage(CumulativeKickoffStats)


# @@protoc_insertion_point(module_scope)
